# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""Definition and mapping of model types to classes"""

from abc import ABCMeta, abstractmethod
import os
import sys

import json
import numpy as np
import pandas as pd
import scipy.sparse as sps
from scipy.interpolate import interpn
from scipy.optimize import fsolve
from scipy.sparse import csc_matrix
from scipy.spatial import distance_matrix
from scipy.stats import truncnorm, norm, gamma, poisson

import connectome_manipulator
from connectome_manipulator import log

P_TH_ABS = 0.01  # Absolute probability threshold
P_TH_REL = 0.1  # Relative probability threshold

N_SYN_PER_CONN_NAME = "n_syn_per_conn"


class AbstractModel(metaclass=ABCMeta):
    """Abstract base class for different types of models."""

    ###########################################################################
    # Abstract properties/methods to be defined in specific model subclasses
    @property
    @abstractmethod
    def param_names(self):
        """Names of model parameters which are part of the model."""

    @property
    @abstractmethod
    def param_defaults(self):
        """Default parameter values as dict. Can be empty if no default values are provided."""

    @property
    @abstractmethod
    def data_names(self):
        """Names of model data frames which are part of the model."""

    @property
    @abstractmethod
    def input_names(self):
        """Names of model inputs which are part of the model."""

    @abstractmethod
    def get_model_output(self, **kwargs):
        """Abstract method for returning model output given its model inputs."""

    @abstractmethod
    def __str__(self):
        """Abstract method for returning a model string describing the model."""

    #
    ###########################################################################

    @staticmethod
    def init_model(model_spec):
        """Model creation from file (within dict) or dict."""
        log.log_assert(
            isinstance(model_spec, dict), "Model specification must be provided as dict!"
        )
        if "file" in model_spec:  # Init from file
            log.log_assert(
                len(model_spec) == 1, "Inconsistent model specification!"
            )  # If file provided, no other entries allowed
            return AbstractModel.model_from_file(model_spec["file"])
        else:  # Init directly from dict
            return AbstractModel.model_from_dict(model_spec)

    @staticmethod
    def model_from_file(model_file):
        """Wrapper function to load model object from file."""
        log.log_assert(os.path.exists(model_file), f'Model file "{model_file}" not found!')
        log.log_assert(
            os.path.splitext(model_file)[1] == ".json",
            'Model file must be of type ".json"!',
        )
        with open(model_file, "r") as f:
            model_dict = json.loads(f.read())
        data_keys = model_dict.pop("data_keys", [])
        data_dict = AbstractModel.load_data_dict(model_file, data_keys)

        return AbstractModel.model_from_dict(model_dict, data_dict)

    @staticmethod
    def load_data_dict(model_file, data_keys):
        """Load supplementary model data (if any) from .h5 data file. [same name and folder as .json file]"""
        if len(data_keys) > 0:
            data_file = os.path.splitext(model_file)[0] + ".h5"
            log.log_assert(os.path.exists(data_file), f'Data file "{data_file}" missing!')
            data_dict = {key: pd.read_hdf(data_file, key) for key in data_keys}
        else:
            data_dict = {}
        return data_dict

    @staticmethod
    def model_from_dict(model_dict, data_dict=None):
        """Wrapper function to create model object from dict."""
        log.log_assert("model" in model_dict, "Model type not found!")
        if data_dict is None:
            data_dict = {}

        model_dict = model_dict.copy()
        model_type = model_dict.pop("model")
        model_class = getattr(sys.modules[__class__.__module__], model_type)  # Get model subclass

        model = model_class(**model_dict, **data_dict)  # Initialize model object

        return model

    def __init__(self, **kwargs):
        """Model initialization from kwargs."""
        self.init_params(kwargs)
        self.init_data(kwargs)

        unused_params = [
            k for k in kwargs if not k.startswith("__")
        ]  # Unused paramters, excluding meta data ('__<name>') that may be included in file
        if len(unused_params) > 0:
            log.warning(f"Unused parameter(s): {set(unused_params)}!")

    def init_params(self, model_dict):
        """Initialize model parameters from dict (removing used keys from dict)."""
        log.log_assert(
            all(p in model_dict or p in self.param_defaults for p in self.param_names),
            f"Missing parameters for model initialization! Must contain initialization for {set(self.param_names) - set(self.param_defaults)}.",
        )
        for p in self.param_names:
            if p in model_dict:
                val = model_dict.pop(p)
            else:  # Use value from defaults
                val = self.param_defaults[p]
            setattr(self, p, val)

    def init_data(self, data_dict):
        """Initialize data frames with supplementary model data from dict (removing used keys from dict)."""
        log.log_assert(
            all(d in data_dict for d in self.data_names),
            f"Missing data for model initialization! Must contain initialization for {set(self.data_names)}.",
        )
        log.log_assert(
            np.all(isinstance(data_dict[d], pd.DataFrame) for d in self.data_names),
            "Model data must be Pandas dataframes!",
        )
        for d in self.data_names:
            setattr(self, d, data_dict.pop(d))

    def apply(self, **kwargs):
        """Main method for applying model, i.e., returning model output given its model inputs.

        [Calls get_model_output() which must be implemented in specific model subclass!]
        """
        #         log.log_assert(
        #             all(inp in kwargs for inp in self.input_names),
        #             f"Missing model inputs! Must contain input values for {set(self.input_names)}.",
        #         )
        inp_dict = {inp: kwargs.pop(inp) for inp in self.input_names}
        if len(kwargs) > 0:
            log.debug(f"Unused input(s): {set(kwargs.keys())}!")
        return self.get_model_output(**inp_dict)

    def get_param_dict(self):
        """Return model parameters as dict."""
        return {p: getattr(self, p) for p in self.param_names}

    def get_data_dict(self):
        """Return data frames with supplementary model data as dict."""
        return {d: getattr(self, d) for d in self.data_names}

    def save_model(self, model_path, model_name):
        """Save model to file: Model dict as .json, model data (if any) as supplementary .h5 data file."""
        model_dict = self.get_param_dict()
        model_file = os.path.join(model_path, model_name + ".json")

        # Save supplementary model data (if any) to .h5 data file
        data_dict = self.get_data_dict()
        log.log_assert(
            np.all(isinstance(v, pd.DataFrame) for k, v in data_dict.items()),
            "Model data must be Pandas dataframes!",
        )
        data_file = os.path.splitext(model_file)[0] + ".h5"
        for idx, (key, df) in enumerate(data_dict.items()):
            df.to_hdf(data_file, key, mode="w" if idx == 0 else "a")
        model_dict["data_keys"] = list(data_dict.keys())

        # Save model dict to .json file
        model_dict["model"] = self.__class__.__name__
        model_dict["__version_info__"] = {
            "connectome_manipulator": connectome_manipulator.__version__,
            "python": sys.version,
            "pandas": pd.__version__,
        }
        with open(model_file, "w") as f:
            f.write(json.dumps(model_dict, indent=2))


# MODEL TEMPLATE #
# class TemplateModel(AbstractModel):
#     """ <Template> model:
#         -Model details...
#     """
#
#     # Names of model inputs, parameters and data frames which are part of this model
#     param_names = [...]
#     param_defaults = {...}
#     data_names = [...]
#     input_names = [...]
#
#     def __init__(self, **kwargs):
#         """Model initialization."""
#         super().__init__(**kwargs)
#
#         # Check parameters
#         log.log_assert(...)
#
#     # <Additional access methods, if needed>
#     ...
#
#     def get_model_output(self, **kwargs):
#         """Description..."""
#         # MUST BE IMPLEMENTED
#         return ...
#
#     def __str__(self):
#         """Return model string describing the model."""
#         model_str = f'{self.__class__.__name__}\n'
#         model_str = model_str + ...
#         return model_str


class PathwayModel(AbstractModel, metaclass=ABCMeta):
    """Abstract model base class for storing model properties per pathway (i.e., for pairs of m-types):

    - Different property values for specific pathways
    - Default property values for any other pathways or properties not specified
    - Actual functionaliy (i.e., how to use these properties) to be implemented in derived class
    """

    # Names of base model inputs and parameters that are part of this model
    pathway_input_names = ["src_type", "tgt_type"]

    def init_params(self, _):
        """Overwritten to avoid setting class variables."""

    def get_param_dict(self):
        """Return model default parameters as dict."""
        return {p: getattr(self, p)[-1, -1] for p in self.param_names}

    @property
    def shorthand(self):
        """Pathway column prefix for this model."""
        return self.__class__.__name__.replace("Model", "").lower()

    def __init__(self, src_type_map=None, tgt_type_map=None, pathway_specs=None, **kwargs):
        """Model initialization."""
        self.property_names = self.param_names
        self.input_names = self.pathway_input_names + self.input_names

        if pathway_specs is not None and (src_type_map is None or tgt_type_map is None):
            raise ValueError("Need to specify both type maps when using pathway specs.")

        self.default_enforce = False
        if src_type_map:
            self.src_type_map = src_type_map
        else:
            self.default_enforce = True
            self.src_type_map = {}
        if tgt_type_map:
            self.tgt_type_map = tgt_type_map
        else:
            self.default_enforce = True
            self.tgt_type_map = {}

        self.default_types = (len(self.src_type_map), len(self.tgt_type_map))
        shape = tuple(s + 1 for s in self.default_types)
        for param in self.param_names:
            if param in kwargs:
                default = kwargs.pop(param)
            else:
                default = self.param_defaults[param]
            matrix = np.full(shape, default)
            colname = "_".join([self.shorthand, param])
            if pathway_specs is not None and colname in pathway_specs.columns:
                col = pathway_specs[colname]
                if ("*", "*") in col:
                    if not np.isnan(default := col.pop(("*", "*"))):
                        matrix[:] = default
                for (src, dst), val in col.items():
                    if not np.isnan(val):
                        i = self.src_type_map[src]
                        j = self.tgt_type_map[dst]
                        matrix[i][j] = val
            setattr(self, param, matrix)

        super().__init__(**kwargs)

    def apply(self, **kwargs):
        """Apply the model.

        Makes sure that the source and destination type values are set according to the
        inputs to the model: if no pathways have been given at construction time, the
        application will always use the default values.
        """
        for way, default in zip(("src", "tgt"), self.default_types):
            way_type = f"{way}_type"
            way_pos = f"{way}_pos"
            if self.default_enforce or way_type not in kwargs:
                log.debug(f"Using default values for {way_type} in {self}")
                if way_type in kwargs:
                    kwargs[way_type] = np.full_like(kwargs[way_type], default)
                elif way_pos in kwargs:
                    kwargs[way_type] = np.full((kwargs[way_pos].shape[0],), default)
                else:
                    kwargs[way_type] = [default]
            elif (types := kwargs.get(way_type, None)) is not None:
                if np.isscalar(types):
                    if way_pos in kwargs:
                        types = np.full((kwargs[way_pos].shape[0],), types)
                    else:
                        types = np.array([types])
                    kwargs[way_type] = types
                kind = np.array(types).dtype.kind
                if kind in ("i", "u"):
                    # No mapping needed for integer indices
                    pass
                elif kind == "U":
                    mapping = getattr(self, f"{way}_type_map")
                    kwargs[way_type] = np.array([mapping[t] for t in types])
                else:
                    raise ValueError(f"Cannot have types '{kind}' for '{way_type}'")
        return super().apply(**kwargs)

    def __str__(self):
        """Return model string describing the model."""
        model_str = f"{self.__class__.__name__}\n"
        model_str += f"  Model properties: {self.property_names}\n"
        if self.default_types[0] > 0 or self.default_types[1] > 0:
            counts = "×".join([str(cnt) for cnt in self.default_types if cnt > 0])
            model_str += f"  Property values for {counts} pathways\n"
        model_str += f"  Default: {self.get_param_dict()}"
        return model_str


class NSynConnModel(PathwayModel):
    """Model for number of synapses per connection for pairs of m-types [generative model]:

    - Synapses per connection drawn from gamma distribution with given mean/std
    - Integer number of synapses larger or equal to one will be returned
    - Different distribution attributes for specific pathways
    - Default distribution attributes for any other pathways not specified
    """

    # Names of model inputs, parameters and data frames which are part of this model
    # (other than the ones inherited from PathwayModel class)
    param_names = ["mean", "std"]
    param_defaults = {"mean": 3.0, "std": 1.5}
    data_names = []
    input_names = []

    def __init__(self, **kwargs):
        """Model initialization."""
        super().__init__(**kwargs)

        log.log_assert(np.all(self.mean > 0.0), "Mean must be larger than zero!")
        log.log_assert(np.all(self.std >= 0.0), "Std cannot be negative!")

    def get_model_output(self, src_type, tgt_type):  # pylint: disable=arguments-differ
        """Draw #syn/conn value for one connection between src_type and tgt_type [seeded through numpy]."""
        # Get distribution attribute values
        distr_mean = self.mean[src_type, tgt_type]
        distr_std = self.std[src_type, tgt_type]

        nsyn = distr_mean
        # Draw number of synapses
        if np.any(sel := distr_std > 0.0):
            nsyn[sel] = np.random.gamma(
                shape=distr_mean[sel] ** 2 / distr_std[sel] ** 2,
                scale=distr_std[sel] ** 2 / distr_mean[sel],
            )

        # Convert type
        nsyn = np.round(np.maximum(1, nsyn)).astype(int)

        return nsyn


class LinDelayModel(PathwayModel):
    """Linear distance-dependent delay model for pairs of m-types [generative model]:

    - Delay mean: delay_mean_coeff_b * distance + delay_mean_coeff_a (linear)
    - Delay std: delay_std (constant)
    - Delay min: delay_min (constant)
    - Different delay attributes for specific pathways
    - Default delay attributes for any other pathways not specified
    """

    # Names of model inputs, parameters and data frames which are part of this model
    # (other than the ones inherited from PathwayModel class)
    param_names = ["delay_mean_coeff_a", "delay_mean_coeff_b", "delay_std", "delay_min"]
    param_defaults = {
        "delay_mean_coeff_a": 0.75,
        "delay_mean_coeff_b": 0.003,
        "delay_std": 0.5,
        "delay_min": 0.2,
    }
    data_names = []
    input_names = ["distance"]

    def __init__(self, **kwargs):
        """Model initialization."""
        super().__init__(**kwargs)

        log.log_assert(np.all(self.delay_std >= 0.0), "Delay std cannot be negative!")
        log.log_assert(np.all(self.delay_min >= 0.0), "Delay min cannot be negative!")

    def get_mean(self, src_type, tgt_type, distance):
        """Returns mean delay as function of distance."""
        return (
            self.delay_mean_coeff_b[src_type, tgt_type] * np.array(distance)
            + self.delay_mean_coeff_a[src_type, tgt_type]
        )

    def get_std(self, src_type, tgt_type, distance):
        """Returns delay std as function of distance."""
        return np.full_like(
            distance, self.delay_std[src_type, tgt_type], dtype=self.delay_std.dtype
        )

    def get_min(self, src_type, tgt_type, distance):
        """Returns min delay as function of distance."""
        return np.full_like(
            distance, self.delay_min[src_type, tgt_type], dtype=self.delay_min.dtype
        )

    def get_model_output(self, src_type, tgt_type, distance):  # pylint: disable=arguments-differ
        """Draw distance-dependent delay values from truncated normal distribution [seeded through numpy]."""
        d_mean = self.get_mean(src_type, tgt_type, distance)
        d_std = self.get_std(src_type, tgt_type, distance)
        d_min = self.get_min(src_type, tgt_type, distance)
        if all(d_std > 0.0):
            return truncnorm.rvs(a=(d_min - d_mean) / d_std, b=np.inf, loc=d_mean, scale=d_std)
        else:
            return np.maximum(d_mean, d_min)  # Deterministic


class PosMapModel(AbstractModel):
    """Position mapping model, mapping one coordinate system to another for a given set of neurons:

    - Mapped neuron position: pos_table.loc[gids] (lookup-table)
    """

    # Names of model inputs, parameters and data frames which are part of this model
    param_names = []
    param_defaults = {}
    data_names = ["pos_table"]
    input_names = ["gids"]

    def get_gids(self):
        """Return GIDs that are mapped within this model."""
        return self.pos_table.index.values

    def get_coord_names(self):
        """Return coordinate names of this model."""
        return list(self.pos_table.columns)

    def get_model_output(self, gids):  # pylint: disable=arguments-differ
        """Return (mapped) neuron positions for a given set of GIDs."""
        return self.pos_table.loc[gids].to_numpy()

    def __str__(self):
        """Return model string describing the model."""
        model_str = f"{self.__class__.__name__}\n"
        model_str = (
            model_str
            + f"  Size: {len(self.get_gids())} GIDs ({np.min(self.get_gids())}..{np.max(self.get_gids())})\n"
        )
        model_str = (
            model_str
            + f'  Outputs: {len(self.get_coord_names())} ({", ".join(self.get_coord_names())})\n'
        )
        model_str = (
            model_str
            + "  Range: "
            + ", ".join(
                [
                    f"{k}: {self.pos_table[k].min():.1f}..{self.pos_table[k].max():.1f}"
                    for k in self.get_coord_names()
                ]
            )
        )
        return model_str


class ConnProbModel(PathwayModel):
    """Generic connection probability model:

    - Implements both Erdos-Renyi and distance-dependent models
    - Returns connection probability for given source/target neuron positions and m-types
    """

    # Names of model inputs, parameters and data frames which are part of this model
    param_names = ["order", "coeff_a", "coeff_b"]
    param_defaults = {"order": 1, "coeff_a": 0.0, "coeff_b": np.nan}
    data_names = []
    input_names = ["src_pos", "tgt_pos"]

    def get_model_output(
        self, src_type, tgt_type, src_pos, tgt_pos
    ):  # pylint: disable=arguments-differ
        """Return pathway-specific connection probabilities <#src x #tgt> for all combinations of source/target neuron positions <#src/#tgt x #dim>."""
        if np.isscalar(src_type):
            src_type = [src_type] * src_pos.shape[0]
        if np.isscalar(tgt_type):
            tgt_type = [tgt_type] * tgt_pos.shape[0]

        p_mat = np.zeros((len(src_type), len(tgt_type)))
        for ti, tt in enumerate(tgt_type):
            # tt is the first index in the Pandas DataFrame
            order = self.order[src_type, tt]
            coeff_a = self.coeff_a[src_type, tt]
            p_mat[:, ti] = coeff_a

            # override higher order values that are more expensive to calculate
            sel = order == 2
            if np.any(sel):
                coeff_a = coeff_a[sel]
                coeff_b = self.coeff_b[src_type, tt][sel]
                distance = np.sqrt(np.sum((src_pos[sel] - tgt_pos[ti]) ** 2, axis=1))
                p_mat[:, ti][sel] = coeff_a * np.exp(-coeff_b * distance)
        return p_mat


class ConnPropsModel(AbstractModel):
    """Connection/synapse properties model for pairs of m-types [generative model]:

    - Connection/synapse property values drawn from given distributions

    NOTE: 'shared_within' flag is used to indicate that same property values are used
          for all synapses within the same connection. Otherwise, property values are
          drawn for all synapses independently from same distribution.

    NOTE: Correlations between properties can be specified using a 'prop_cov' dict with
          'props' (list of correlated property names) and 'cov' (covariance matrices
          by pathway)
    """

    # Names of model inputs, parameters and data frames which are part of this model
    param_names = ["src_types", "tgt_types", "prop_stats", "prop_cov"]
    param_defaults = {"prop_cov": {}}
    data_names = []
    input_names = ["src_type", "tgt_type", "n_syn"]
    # Notes:
    # n_syn is optional; it must be provided if no N_SYN_PER_CONN_NAME property is specified
    # If N_SYN_PER_CONN_NAME property is specified but n_syn is provided, n_syn will be used

    # Required attributes for given distributions
    distribution_attributes = {
        "constant": ["mean"],
        "normal": ["mean", "std"],
        "truncnorm": ["norm_loc", "norm_scale", "min", "max"],
        "gamma": ["mean", "std"],
        "poisson": ["mean"],
        "ztpoisson": ["mean"],
        "discrete": ["val", "p"],
        "zero": [],
    }
    # Notes:
    # constant: "mean" corresponds to the constant value
    # truncnorm: "norm_loc"/"norm_scale" are center location (mean) and scale (standard deviation) of the underlying (non-truncated) normal distribution
    # ztpoisson: Zero-truncated poisson distribution
    # discrete: "val"/"p" are discrete values and probabilities of the discrete distribution
    # zero: Empty distribution always returning zero, to model unused parameters

    def __init__(self, **kwargs):
        """Model initialization."""
        super().__init__(**kwargs)

        def dict_conv(data):
            """Recursively convert numpy to basic data types, to have a clean JSON file"""
            if isinstance(data, dict):
                return {k: dict_conv(v) for k, v in data.items()}
            elif isinstance(data, (list, tuple)):
                return [dict_conv(d) for d in data]
            elif hasattr(data, "tolist"):  # Convert numpy types
                return data.tolist()
            else:
                return data

        # Check & convert parameters
        # pylint: disable=access-member-before-definition
        log.log_assert(isinstance(self.prop_stats, dict), '"prop_stats" dictionary required!')
        self.prop_stats = dict_conv(self.prop_stats)  # Convert dict to basic data types
        self.prop_names = list(self.prop_stats.keys())
        if N_SYN_PER_CONN_NAME in self.prop_names:
            self.has_nsynconn = True
        else:
            self.has_nsynconn = False
        log.log_assert(
            all(isinstance(self.prop_stats[p], dict) for p in self.prop_names),
            "Property statistics dictionary required!",
        )
        log.log_assert(
            all(
                all(np.isin(self.src_types, list(self.prop_stats[p].keys())))
                for p in self.prop_names
            ),
            f"Source type statistics missing! self.src_types:{self.src_types} not in {self.prop_names}",
        )
        log.log_assert(
            all(
                all(isinstance(self.prop_stats[p][src], dict) for p in self.prop_names)
                for src in self.src_types
            ),
            "Property statistics dictionary required!",
        )
        log.log_assert(
            all(
                all(
                    all(np.isin(self.tgt_types, list(self.prop_stats[p][src].keys())))
                    for p in self.prop_names
                )
                for src in self.src_types
            ),
            "Target type statistics missing!",
        )
        log.log_assert(
            all(
                all(
                    all("type" in self.prop_stats[p][src][tgt].keys() for p in self.prop_names)
                    for src in self.src_types
                )
                for tgt in self.tgt_types
            ),
            "Distribution type missing!",
        )
        log.log_assert(
            all(
                all(
                    all(
                        all(
                            np.isin(
                                self.distribution_attributes[self.prop_stats[p][src][tgt]["type"]],
                                list(self.prop_stats[p][src][tgt].keys()),
                            )
                        )
                        for p in self.prop_names
                    )
                    for src in self.src_types
                )
                for tgt in self.tgt_types
            ),
            f"Distribution attributes missing (required: {self.distribution_attributes})!",
        )
        log.log_assert(
            all(
                all(
                    all(
                        all(
                            len(self.prop_stats[p][src][tgt][a]) > 0
                            for a in self.distribution_attributes[
                                self.prop_stats[p][src][tgt]["type"]
                            ]
                            if hasattr(self.prop_stats[p][src][tgt][a], "__iter__")
                        )
                        for p in self.prop_names
                    )
                    for src in self.src_types
                )
                for tgt in self.tgt_types
            ),
            f"Distribution attribute(s) empty (required: {self.distribution_attributes})!",
        )
        log.log_assert(
            all(
                all(
                    all(
                        np.isclose(np.sum(self.prop_stats[p][src][tgt]["p"]), 1.0)
                        for p in self.prop_names
                        if "p" in self.prop_stats[p][src][tgt].keys()
                    )
                    for src in self.src_types
                )
                for tgt in self.tgt_types
            ),
            'Probability attribute "p" does not sum to 1.0!',
        )
        log.log_assert(
            all(
                all(
                    all(
                        len(self.prop_stats[p][src][tgt]["p"])
                        == len(self.prop_stats[p][src][tgt]["val"])
                        for p in self.prop_names
                        if np.all(np.isin(["p", "val"], list(self.prop_stats[p][src][tgt].keys())))
                    )
                    for src in self.src_types
                )
                for tgt in self.tgt_types
            ),
            'Probability attribute "p" does not match length of corresponding "val"!',
        )
        log.log_assert(
            all(
                all(
                    all(
                        self.prop_stats[p][src][tgt]["lower_bound"]
                        <= self.prop_stats[p][src][tgt]["upper_bound"]
                        for p in self.prop_names
                        if "lower_bound" in self.prop_stats[p][src][tgt].keys()
                        and "upper_bound" in self.prop_stats[p][src][tgt].keys()
                    )
                    for src in self.src_types
                )
                for tgt in self.tgt_types
            ),
            "Data bounds error!",
        )
        log.log_assert(isinstance(self.prop_cov, dict), '"prop_cov" dictionary required!')
        self.prop_cov = dict_conv(self.prop_cov)  # Convert dict to basic data types
        if len(self.prop_cov) > 0:
            log.log_assert(
                "props" in self.prop_cov and "cov" in self.prop_cov,
                'Property COV dict must contain "props" and "cov" keys!',
            )
            log.log_assert(len(self.prop_cov["props"]) > 0, 'Property COV "props" names missing!')
            log.log_assert(
                all(np.isin(self.prop_cov["props"], self.prop_names)),
                'Property COV "props" name error!',
            )
            log.log_assert(
                N_SYN_PER_CONN_NAME not in self.prop_cov["props"],
                f'Property COV for "{N_SYN_PER_CONN_NAME}" not supported!',
            )
            for src in self.src_types:
                log.log_assert(
                    src in self.prop_cov["cov"],
                    f'Property COV source mtype "{src}" missing!',
                )
                for tgt in self.tgt_types:
                    log.log_assert(
                        tgt in self.prop_cov["cov"][src],
                        f'Property COV target mtype "{tgt}" for source "{src}" missing!',
                    )
                    log.log_assert(
                        all(
                            np.array(np.array(self.prop_cov["cov"][src][tgt]).shape)
                            == len(self.prop_cov["props"])
                        ),
                        "Propert COV matrix size mismatch!",
                    )
                    log.log_assert(
                        np.all(np.diag(self.prop_cov["cov"][src][tgt]) == 1.0),
                        "Property COV matrices must contain ones in the diagonal!",
                    )
                    log.log_assert(
                        np.array_equal(
                            self.prop_cov["cov"][src][tgt],
                            np.array(self.prop_cov["cov"][src][tgt]).T,
                        ),
                        "Property COV matrices must be symmetric!",
                    )
            self.prop_cov_names = self.prop_cov["props"]
            self.prop_cov_mat = self.prop_cov["cov"]
        else:
            self.prop_cov_names = []
            self.prop_cov_mat = None

    def get_prop_names(self):
        """Return list of connection/synapse property names."""
        return self.prop_names

    def get_prop_cov_names(self):
        """Return list of correlated property names."""
        return self.prop_cov_names

    def get_prop_cov_mat(self, src_type, tgt_type):
        """Return covariance matrix of correlated properties."""
        return np.array(self.prop_cov_mat[src_type][tgt_type])

    def get_src_types(self):
        """Return list source (pre-synaptic) m-types."""
        return self.src_types

    def get_tgt_types(self):
        """Return list target (post-synaptic) m-types."""
        return self.tgt_types

    def get_distr_props(self, prop_name, src_type, tgt_type):
        """Return distribution type & properties (mean, std, ...)."""
        return self.prop_stats[prop_name][src_type][tgt_type]

    @staticmethod
    def zero_truncated_poisson(lam, size=1):
        """Draw value(s) from zero-truncated poisson distribution."""
        u = np.random.uniform(np.exp(-lam), 1, size=size)
        t = -np.log(u)
        return 1 + np.random.poisson(lam - t)

    @staticmethod
    def compute_ztpoisson_lambda(mean):
        """Compute lambda of zero-truncated poission distribution corresponding to given mean (numerically)."""

        def fct(lam, mn):
            return lam / (1 - np.exp(-lam)) - mn

        lam = fsolve(fct, mean, mean)[0]
        lam = np.maximum(lam, 0.0)
        return lam

    @staticmethod
    def draw_from_distribution(distr_spec, size=1):
        """Draw value(s) from given distribution"""
        distr_type = distr_spec.get("type")
        if distr_type == "constant":
            distr_val = distr_spec.get("mean")
            log.log_assert(
                distr_val is not None,
                "Distribution parameter missing (required: mean)!",
            )
            drawn_values = np.full(size, distr_val)
        elif distr_type == "normal":
            distr_mean = distr_spec.get("mean")
            distr_std = distr_spec.get("std")
            log.log_assert(
                distr_mean is not None and distr_std is not None,
                "Distribution parameter missing (required: mean/std)!",
            )
            drawn_values = np.random.normal(loc=distr_mean, scale=distr_std, size=size)
        elif distr_type == "truncnorm":
            distr_loc = distr_spec.get("norm_loc")
            distr_scale = distr_spec.get("norm_scale")
            distr_min = distr_spec.get("min")
            distr_max = distr_spec.get("max")
            log.log_assert(
                distr_loc is not None
                and distr_scale is not None
                and distr_min is not None
                and distr_max is not None,
                "Distribution parameters missing (required: norm_loc/norm_scale/min/max)!",
            )
            log.log_assert(distr_min <= distr_max, "Range error (truncnorm)!")
            if distr_scale > 0.0:
                drawn_values = truncnorm(
                    a=(distr_min - distr_loc) / distr_scale,
                    b=(distr_max - distr_loc) / distr_scale,
                    loc=distr_loc,
                    scale=distr_scale,
                ).rvs(size=size)
            else:
                drawn_values = np.clip(np.full(size, distr_loc), distr_min, distr_max)

        elif distr_type == "gamma":
            distr_mean = distr_spec.get("mean")
            distr_std = distr_spec.get("std")
            log.log_assert(
                distr_mean is not None and distr_std is not None,
                "Distribution parameter missing (required: mean/std)!",
            )
            log.log_assert(distr_mean > 0.0 and distr_std >= 0.0, "Range error (gamma)!")
            if distr_std > 0.0:
                drawn_values = np.random.gamma(
                    shape=distr_mean**2 / distr_std**2,
                    scale=distr_std**2 / distr_mean,
                    size=size,
                )
            else:
                drawn_values = np.full(size, distr_mean)
        elif distr_type == "poisson":
            distr_mean = distr_spec.get("mean")
            log.log_assert(
                distr_mean is not None, "Distribution parameter missing (required: mean)!"
            )
            log.log_assert(distr_mean >= 0.0, "Range error (poisson)!")
            drawn_values = np.random.poisson(lam=distr_mean, size=size)
        elif distr_type == "ztpoisson":
            distr_mean = distr_spec.get("mean")
            log.log_assert(
                distr_mean is not None, "Distribution parameter missing (required: mean)!"
            )
            log.log_assert(distr_mean >= 1.0, "Range error (zero-truncated poisson)!")

            # Determine lambda corresponding to given mean...
            lam = ConnPropsModel.compute_ztpoisson_lambda(distr_mean)
            # ...and draw values from zero-truncated poisson distribition
            drawn_values = ConnPropsModel.zero_truncated_poisson(lam=lam, size=size)
        elif distr_type == "discrete":
            distr_val = distr_spec.get("val")
            distr_p = distr_spec.get("p")
            log.log_assert(
                distr_val is not None
                and distr_p is not None
                and not np.isscalar(distr_val)
                and not np.isscalar(distr_p)
                and len(distr_val) == len(distr_p),
                "Distribution parameters error or missing (required: list-like val/p of same length)!",
            )
            drawn_values = np.random.choice(distr_val, size=size, p=distr_p)
        elif distr_type == "zero":
            drawn_values = np.full(size, 0.0)
        else:
            log.log_assert(False, f'Distribution type "{distr_type}" not supported!')
        return drawn_values  # pylint: disable=E0606

    def draw(self, prop_name, src_type, tgt_type, size=1):
        """Draw value(s) for given property name of a single connection

        (or multiple connections, if prop_name==N_SYN_PER_CONN_NAME)
        """
        stats_dict = self.prop_stats.get(prop_name)[src_type][tgt_type]
        if prop_name == N_SYN_PER_CONN_NAME:  # Draw <size> N_SYN_PER_CONN_NAME value(s)
            drawn_values = np.maximum(
                np.round(self.draw_from_distribution(stats_dict, size)).astype(int),
                1,
            )  # At least one synapse/connection, otherwise no connection!!
        else:
            shared_within = stats_dict.get("shared_within", True)
            if shared_within:  # Same property value for all synapses within connection
                val = self.draw_from_distribution(stats_dict, 1)  # Draw a single value...
                drawn_values = np.full(size, val)  # ...and reuse in all synapses
            else:  # Redraw property values independently for all synapses within connection
                drawn_values = self.draw_from_distribution(stats_dict, size)

        # Apply upper/lower bounds (optional)
        lower_bound = stats_dict.get("lower_bound")
        upper_bound = stats_dict.get("upper_bound")
        if lower_bound is not None:
            drawn_values = np.maximum(drawn_values, lower_bound)
        if upper_bound is not None:
            drawn_values = np.minimum(drawn_values, upper_bound)

        # Set data type (optional)
        data_type = stats_dict.get("dtype")
        if data_type is not None:
            if data_type == "int":
                drawn_values = np.round(drawn_values)
            drawn_values = drawn_values.astype(data_type)

        return drawn_values

    @staticmethod
    def remap_distribution(values, distr_spec):
        """Re-map values from standard normal disrtibution to another distributions."""
        sf_values = 1 - norm.cdf(values)
        remapped_values = None
        distr_type = distr_spec.get("type")
        if distr_type == "normal":
            distr_mean = distr_spec.get("mean")
            distr_std = distr_spec.get("std")
            remapped_values = norm.isf(sf_values, loc=distr_mean, scale=distr_std)
        elif distr_type == "truncnorm":
            distr_loc = distr_spec.get("norm_loc")
            distr_scale = distr_spec.get("norm_scale")
            distr_min = distr_spec.get("min")
            distr_max = distr_spec.get("max")
            log.log_assert(
                distr_scale > 0.0, 'Truncnorm remapping error: "norm_scale" cannot be zero!'
            )
            a = (distr_min - distr_loc) / distr_scale
            b = (distr_max - distr_loc) / distr_scale
            remapped_values = truncnorm.isf(sf_values, a=a, b=b, loc=distr_loc, scale=distr_scale)
        elif distr_type == "gamma":
            distr_mean = distr_spec.get("mean")
            distr_std = distr_spec.get("std")
            log.log_assert(
                distr_mean > 0.0 and distr_std > 0.0,
                'Gamma remapping error: "mean" and "std" cannot be zero!',
            )
            shape = distr_mean**2 / distr_std**2
            scale = distr_std**2 / distr_mean
            remapped_values = gamma.isf(sf_values, a=shape, scale=scale)
        elif distr_type == "poisson":
            distr_mean = distr_spec.get("mean")
            remapped_values = poisson.isf(sf_values, mu=distr_mean)
        else:
            log.log_assert(False, f'Remapping not supported for distribution type "{distr_type}"!')

        # Apply upper/lower bounds (optional)
        lower_bound = distr_spec.get("lower_bound")
        upper_bound = distr_spec.get("upper_bound")
        if lower_bound is not None:
            remapped_values = np.maximum(remapped_values, lower_bound)
        if upper_bound is not None:
            remapped_values = np.minimum(remapped_values, upper_bound)

        # Set data type (optional)
        data_type = distr_spec.get("dtype")
        if data_type is not None:
            if data_type == "int":
                remapped_values = np.round(remapped_values)
            remapped_values = remapped_values.astype(data_type)

        return remapped_values

    def draw_cov(self, src_type, tgt_type, size=1):
        """Draw correlated property values of a single connection.

        (As in Chindemi et al. (2022) "A calcium-based plasticity model for
        predicting long-term potentiation and depression in the neocortex")
        """
        if self.prop_cov_mat is None:
            log.warning("No correlated properties!")
            return np.zeros((size, 0))

        stats_dicts = [self.prop_stats.get(p)[src_type][tgt_type] for p in self.prop_cov_names]
        shared_within = [_sdict.get("shared_within", True) for _sdict in stats_dicts]
        log.log_assert(
            np.all(np.array(shared_within) == shared_within[0]),
            'Inconsistent "shared_within" value among correlated properties!',
        )
        shared_within = shared_within[0]

        # Draw from multivariate Gaussian
        if shared_within:  # Same property value for all synapses within connection
            val = np.random.multivariate_normal(
                np.zeros(len(self.prop_cov_names)), self.prop_cov_mat[src_type][tgt_type], size=1
            )
            drawn_values = np.repeat(val, size, axis=0)  # Reuse same value in all synapses
        else:  # Redraw property values independently for all synapses within connection
            drawn_values = np.random.multivariate_normal(
                np.zeros(len(self.prop_cov_names)), self.prop_cov_mat[src_type][tgt_type], size=size
            )

        # Re-map Gaussian to respective marginal distributions
        remapped_values = [
            self.remap_distribution(drawn_values[:, [_i]], _sdict)
            for _i, _sdict in enumerate(stats_dicts)
        ]

        return remapped_values

    def get_model_output(self, src_type, tgt_type, n_syn=None):  # pylint: disable=arguments-differ
        """Draw property values for one connection between src_type and tgt_type, returning a dataframe [seeded through numpy]."""
        syn_props = [
            p for p in self.prop_names if p != N_SYN_PER_CONN_NAME and p not in self.prop_cov_names
        ]
        if n_syn is None:
            log.log_assert(self.has_nsynconn, f'"{N_SYN_PER_CONN_NAME}" missing')
            n_syn = self.draw(N_SYN_PER_CONN_NAME, src_type, tgt_type, 1)[0]
        else:
            log.log_assert(n_syn > 0, '"n_syn" must be at least 1!')

        df = pd.DataFrame([], index=range(n_syn), columns=syn_props)
        for p in syn_props:
            df[p] = self.draw(p, src_type, tgt_type, n_syn)

        if len(self.prop_cov_names) > 0:
            # Draw correlated property values
            corr_vals = self.draw_cov(src_type, tgt_type, n_syn)
            for i, p in enumerate(self.prop_cov_names):
                df[p] = corr_vals[i]
        return df

    def apply(self, **kwargs):
        """Apply the model. Overwrite to set default."""
        if "n_syn" not in kwargs:
            kwargs["n_syn"] = None
        return super().apply(**kwargs)

    def __str__(self):
        """Return model string describing the model."""
        distr_types = {
            p: "/".join(
                np.unique(
                    [
                        [self.prop_stats[p][src][tgt]["type"] for src in self.src_types]
                        for tgt in self.tgt_types
                    ]
                )
            )
            for p in self.prop_names
        }  # Extract distribution types
        distr_dtypes = {
            p: "/".join(
                np.unique(
                    [
                        [self.prop_stats[p][src][tgt].get("dtype", "") for src in self.src_types]
                        for tgt in self.tgt_types
                    ]
                )
            )
            for p in self.prop_names
        }  # Extract distribution data types
        model_str = f"{self.__class__.__name__}"
        model_str = model_str + f" ({'with' if self.has_nsynconn else 'w/o'} #syn/conn)\n"
        model_str = (
            model_str
            + f"  Connection/synapse property distributions between {len(self.src_types)}x{len(self.tgt_types)} M-types:\n"
        )
        model_str = (
            model_str
            + "  "
            + "; ".join(
                [
                    f"{p} <{distr_types[p]}"
                    + (f" ({distr_dtypes[p]})" if distr_dtypes[p] else "")
                    + ">"
                    for p in self.prop_names
                ]
            )
        )
        if len(self.prop_cov_names) > 0:
            model_str = model_str + f"\n  Correlated properties: {', '.join(self.prop_cov_names)}"
        return model_str


class ConnProb1stOrderModel(AbstractModel):
    """1st order connection probability model (Erdos-Renyi):

    - Returns (constant) connection probability for given source/target neuron positions
    """

    # Names of model inputs, parameters and data frames which are part of this model
    param_names = ["p_conn"]
    param_defaults = {"p_conn": 0.0}
    data_names = []
    input_names = ["src_pos", "tgt_pos"]

    def __init__(self, **kwargs):
        """Model initialization."""
        super().__init__(**kwargs)

        # Check parameters
        if np.all(np.isnan([getattr(self, p) for p in self.param_names])):
            log.warning("Empty/invalid model!")
        else:
            log.log_assert(
                0.0 <= self.p_conn <= 1.0, "Connection probability must be between 0 and 1!"
            )

    def get_conn_prob(self):
        """Return (constant) connection probability."""
        return self.p_conn

    def get_model_output(self, src_pos, tgt_pos):  # pylint: disable=arguments-differ
        """Return (constant) connection probabilities <#src x #tgt> for all combinations of source/target neuron positions <#src/#tgt x #dim>."""
        #         log.log_assert(
        #             src_pos.shape[1] == tgt_pos.shape[1],
        #             "Dimension mismatch of source/target neuron positions!",
        #         )
        return np.full((src_pos.shape[0], tgt_pos.shape[0]), self.get_conn_prob())

    def __str__(self):
        """Return model string describing the model."""
        model_str = f"{self.__class__.__name__}\n"
        model_str = model_str + f"  p_conn() = {self.p_conn:.3f} (constant)"
        return model_str


class ConnProb2ndOrderExpModel(AbstractModel):
    """2nd order connection probability model (exponential distance-dependent):

    - Returns (distance-dependent) connection probabilities for given source/target neuron positions
    """

    # Names of model inputs, parameters and data frames which are part of this model
    param_names = ["scale", "exponent"]
    param_defaults = {"scale": 0.0, "exponent": 0.0}
    data_names = []
    input_names = ["src_pos", "tgt_pos"]

    def __init__(self, **kwargs):
        """Model initialization."""
        super().__init__(**kwargs)

        # Check parameters
        if np.all(np.isnan([getattr(self, p) for p in self.param_names])):
            log.warning("Empty/invalid model!")
        else:
            log.log_assert(0.0 <= self.scale <= 1.0, '"Scale" must be between 0 and 1!')
            log.log_assert(self.exponent >= 0.0, '"Exponent" must be non-negative!')

    @staticmethod
    def exp_fct(distance, scale, exponent):
        """Distance-dependent exponential probability function."""
        return scale * np.exp(-exponent * np.array(distance))

    def get_conn_prob(self, distance):
        """Return (distance-dependent) connection probability."""
        return self.exp_fct(distance, self.scale, self.exponent)

    @staticmethod
    def compute_dist_matrix(src_pos, tgt_pos):
        """Compute distance matrix between pairs of neurons."""
        dist_mat = distance_matrix(src_pos, tgt_pos)
        dist_mat[dist_mat == 0.0] = np.nan  # Exclude autaptic connections
        return dist_mat

    def get_model_output(self, src_pos, tgt_pos):  # pylint: disable=arguments-differ
        """Return (distance-dependent) connection probabilities <#src x #tgt> for all combinations of source/target neuron positions <#src/#tgt x #dim>."""
        #         log.log_assert(
        #             src_pos.shape[1] == tgt_pos.shape[1],
        #             "Dimension mismatch of source/target neuron positions!",
        #         )
        dist_mat = self.compute_dist_matrix(src_pos, tgt_pos)
        return self.get_conn_prob(dist_mat)

    def __str__(self):
        """Return model string describing the model."""
        model_str = f"{self.__class__.__name__}\n"
        model_str = model_str + f"  p_conn(d) = {self.scale:.3f} * exp(-{self.exponent:.3f} * d)\n"
        model_str = model_str + "  d...distance"
        return model_str


class ConnProb2ndOrderComplexExpModel(AbstractModel):
    """2nd order connection probability model (complex exponential distance-dependent),

    based on a complex (proximal) exponential and a simple (distal) exponential function:
    - Returns (distance-dependent) connection probabilities for given source/target neuron positions
    """

    # Names of model inputs, parameters and data frames which are part of this model
    param_names = ["prox_scale", "prox_exp", "prox_exp_pow", "dist_scale", "dist_exp"]
    param_defaults = {}
    data_names = []
    input_names = ["src_pos", "tgt_pos"]

    def __init__(self, **kwargs):
        """Model initialization."""
        super().__init__(**kwargs)

        # Check parameters
        if np.all(np.isnan([getattr(self, p) for p in self.param_names])):
            log.warning("Empty/invalid model!")
        else:
            log.log_assert(0.0 <= self.prox_scale <= 1.0, '"prox_scale" must be between 0 and 1!')
            log.log_assert(self.prox_exp >= 0.0, '"prox_exp" must be non-negative!')
            log.log_assert(self.prox_exp_pow >= 0.0, '"prox_exp_pow" must be non-negative!')
            log.log_assert(0.0 <= self.dist_scale <= 1.0, '"dist_scale" must be between 0 and 1!')
            log.log_assert(self.dist_exp >= 0.0, '"dist_exp" must be non-negative!')
            test_distance = 1000.0
            if self.exp_fct(test_distance, 1.0, self.prox_exp, self.prox_exp_pow) >= self.exp_fct(
                test_distance, 1.0, self.dist_exp, 1.0
            ):
                log.warning(
                    f"Proximal exponential decays slower than distal exponential ({self.get_param_dict()})!"
                )

    @staticmethod
    def exp_fct(distance, scale, exponent, exp_power=1.0):
        """Distance-dependent (complex) exponential probability function."""
        return scale * np.exp(-exponent * np.array(distance) ** exp_power)

    def get_conn_prob(self, distance):
        """Return (distance-dependent) connection probability."""
        return self.exp_fct(
            distance, self.prox_scale, self.prox_exp, self.prox_exp_pow
        ) + self.exp_fct(distance, self.dist_scale, self.dist_exp, 1.0)

    @staticmethod
    def compute_dist_matrix(src_pos, tgt_pos):
        """Compute distance matrix between pairs of neurons."""
        dist_mat = distance_matrix(src_pos, tgt_pos)
        dist_mat[dist_mat == 0.0] = np.nan  # Exclude autaptic connections
        return dist_mat

    def get_model_output(self, src_pos, tgt_pos):  # pylint: disable=arguments-differ
        """Return (distance-dependent) connection probabilities <#src x #tgt> for all combinations of source/target neuron positions <#src/#tgt x #dim>."""
        #         log.log_assert(
        #             src_pos.shape[1] == tgt_pos.shape[1],
        #             "Dimension mismatch of source/target neuron positions!",
        #         )
        dist_mat = self.compute_dist_matrix(src_pos, tgt_pos)
        return self.get_conn_prob(dist_mat)

    def __str__(self):
        """Return model string describing the model."""
        model_str = f"{self.__class__.__name__}\n"
        model_str = (
            model_str
            + f"  p_conn(d) = {self.prox_scale:.3f} * exp(-{self.prox_exp:.6f} * d^{self.prox_exp_pow:.3f}) + {self.dist_scale:.3f} * exp(-{self.dist_exp:.3f} * d)\n"
        )
        model_str = model_str + "  d...distance"
        return model_str


class ConnProb3rdOrderExpModel(AbstractModel):
    """3rd order connection probability model (bipolar exponential distance-dependent):

    - Returns (bipolar distance-dependent) connection probabilities for given source/target neuron positions
    """

    # Names of model inputs, parameters and data frames which are part of this model
    param_names = ["scale_P", "scale_N", "exponent_P", "exponent_N", "bip_coord"]
    param_defaults = {}
    data_names = []
    input_names = ["src_pos", "tgt_pos"]

    def __init__(self, **kwargs):
        """Model initialization."""
        super().__init__(**kwargs)

        # Check parameters
        if np.all(
            np.isnan([getattr(self, p) for p in np.setdiff1d(self.param_names, "bip_coord")])
        ):
            log.warning("Empty/invalid model!")
        else:
            log.log_assert(
                0.0 <= self.scale_P <= 1.0 and 0.0 <= self.scale_N <= 1.0,
                '"Scale" must be between 0 and 1!',
            )
            log.log_assert(
                self.exponent_P >= 0.0 and self.exponent_N >= 0.0,
                '"Exponent" must be non-negative!',
            )
            log.log_assert(
                isinstance(self.bip_coord, int) and 0 <= self.bip_coord <= 2,
                'Bipolar coordinate "bip_coord" out of range!',
            )

    @staticmethod
    def exp_fct(distance, scale, exponent):
        """Distance-dependent exponential probability function."""
        return scale * np.exp(-exponent * np.array(distance))

    def get_conn_prob(self, distance, bip):
        """Return (bipolar distance-dependent) connection probability."""
        p_conn_N = self.exp_fct(distance, self.scale_N, self.exponent_N)
        p_conn_P = self.exp_fct(distance, self.scale_P, self.exponent_P)
        p_conn = np.select(
            [np.array(bip) < 0.0, np.array(bip) > 0.0],
            [p_conn_N, p_conn_P],
            default=0.5 * (p_conn_N + p_conn_P),
        )
        return p_conn

    @staticmethod
    def compute_dist_matrix(src_pos, tgt_pos):
        """Compute distance matrix between pairs of neurons."""
        dist_mat = distance_matrix(src_pos, tgt_pos)
        dist_mat[dist_mat == 0.0] = np.nan  # Exclude autaptic connections
        return dist_mat

    @staticmethod
    def compute_bip_matrix(src_pos, tgt_pos, bip_coord):
        """Computes bipolar matrix between pairs of neurons along specified coordinate axis (default: 2..z-axis),

        defined as sign of target (POST-synaptic) minus source (PRE-synaptic) coordinate value
        (i.e., POST-synaptic neuron below (delta < 0) or above (delta > 0) PRE-synaptic neuron assuming
         axis values increasing from lower to upper layers)
        """
        bip_mat = np.sign(
            np.diff(
                np.meshgrid(src_pos[:, bip_coord], tgt_pos[:, bip_coord], indexing="ij"), axis=0
            )[0, :, :]
        )  # Bipolar distinction based on difference in specified coordinate
        return bip_mat

    def get_model_output(self, src_pos, tgt_pos):  # pylint: disable=arguments-differ
        """Return (bipolar distance-dependent) connection probabilities <#src x #tgt> for all combinations of source/target neuron positions <#src/#tgt x #dim>."""
        #         log.log_assert(
        #             src_pos.shape[1] == tgt_pos.shape[1],
        #             "Dimension mismatch of source/target neuron positions!",
        #         )
        dist_mat = self.compute_dist_matrix(src_pos, tgt_pos)
        bip_mat = self.compute_bip_matrix(src_pos, tgt_pos, self.bip_coord)
        return self.get_conn_prob(dist_mat, bip_mat)

    def __str__(self):
        """Return model string describing the model."""
        model_str = f"{self.__class__.__name__}\n"
        model_str = (
            model_str
            + f"  p_conn(d, delta) = {self.scale_N:.3f} * exp(-{self.exponent_N:.3f} * d) if delta < 0\n"
        )
        model_str = (
            model_str
            + f"                     {self.scale_P:.3f} * exp(-{self.exponent_P:.3f} * d) if delta > 0\n"
        )
        model_str = model_str + "                     AVERAGE OF BOTH MODELS  if delta == 0\n"
        model_str = (
            model_str
            + f"  d...distance, delta...difference (tgt minus src) in coordinate {self.bip_coord}"
        )
        return model_str


class ConnProb3rdOrderComplexExpModel(AbstractModel):
    """3rd order connection probability model (bipolar complex exponential distance-dependent),

    based on a complex (proximal) exponential and a simple (distal) exponential function
    - Returns (bipolar distance-dependent) connection probabilities for given source/target neuron positions
    """

    # Names of model inputs, parameters and data frames which are part of this model
    param_names = [
        "prox_scale_P",
        "prox_scale_N",
        "prox_exp_P",
        "prox_exp_N",
        "prox_exp_pow_P",
        "prox_exp_pow_N",
        "dist_scale_P",
        "dist_scale_N",
        "dist_exp_P",
        "dist_exp_N",
        "bip_coord",
    ]
    param_defaults = {}
    data_names = []
    input_names = ["src_pos", "tgt_pos"]

    def __init__(self, **kwargs):
        """Model initialization."""
        super().__init__(**kwargs)

        # Check parameters
        if np.all(
            np.isnan([getattr(self, p) for p in np.setdiff1d(self.param_names, "bip_coord")])
        ):
            log.warning("Empty/invalid model!")
        else:
            log.log_assert(
                0.0 <= self.prox_scale_P <= 1.0 and 0.0 <= self.prox_scale_N <= 1.0,
                '"prox_scale_P/N" must be between 0 and 1!',
            )
            log.log_assert(
                self.prox_exp_P >= 0.0 and self.prox_exp_N >= 0.0,
                '"prox_exp_P/N" must be non-negative!',
            )
            log.log_assert(
                self.prox_exp_pow_P >= 0.0 and self.prox_exp_pow_N >= 0.0,
                '"prox_exp_pow_P/N" must be non-negative!',
            )
            log.log_assert(
                0.0 <= self.dist_scale_P <= 1.0 and 0.0 <= self.dist_scale_N <= 1.0,
                '"dist_scale_P/N" must be between 0 and 1!',
            )
            log.log_assert(
                self.dist_exp_P >= 0.0 and self.dist_exp_N >= 0.0,
                '"dist_exp_P/N" must be non-negative!',
            )
            log.log_assert(
                isinstance(self.bip_coord, int) and 0 <= self.bip_coord <= 2,
                'Bipolar coordinate "bip_coord" out of range!',
            )
            test_distance = 1000.0
            if self.exp_fct(
                test_distance, 1.0, self.prox_exp_P, self.prox_exp_pow_P
            ) >= self.exp_fct(test_distance, 1.0, self.dist_exp_P, 1.0):
                log.warning(
                    f"Proximal (P) exponential decays slower than distal (P) exponential ({self.get_param_dict()})!"
                )
            if self.exp_fct(
                test_distance, 1.0, self.prox_exp_N, self.prox_exp_pow_N
            ) >= self.exp_fct(test_distance, 1.0, self.dist_exp_N, 1.0):
                log.warning(
                    f"Proximal (N) exponential decays slower than distal (N) exponential ({self.get_param_dict()})!"
                )

    @staticmethod
    def exp_fct(distance, scale, exponent, exp_power=1.0):
        """Distance-dependent (complex) exponential probability function."""
        return scale * np.exp(-exponent * np.array(distance) ** exp_power)

    def get_conn_prob(self, distance, bip):
        """Return (bipolar distance-dependent) connection probability."""
        p_conn_N = self.exp_fct(
            distance, self.prox_scale_N, self.prox_exp_N, self.prox_exp_pow_N
        ) + self.exp_fct(distance, self.dist_scale_N, self.dist_exp_N, 1.0)
        p_conn_P = self.exp_fct(
            distance, self.prox_scale_P, self.prox_exp_P, self.prox_exp_pow_P
        ) + self.exp_fct(distance, self.dist_scale_P, self.dist_exp_P, 1.0)
        p_conn = np.select(
            [np.array(bip) < 0.0, np.array(bip) > 0.0],
            [p_conn_N, p_conn_P],
            default=0.5 * (p_conn_N + p_conn_P),
        )
        return p_conn

    @staticmethod
    def compute_dist_matrix(src_pos, tgt_pos):
        """Compute distance matrix between pairs of neurons."""
        dist_mat = distance_matrix(src_pos, tgt_pos)
        dist_mat[dist_mat == 0.0] = np.nan  # Exclude autaptic connections
        return dist_mat

    @staticmethod
    def compute_bip_matrix(src_pos, tgt_pos, bip_coord):
        """Computes bipolar matrix between pairs of neurons along specified coordinate axis (default: 2..z-axis),

        defined as sign of target (POST-synaptic) minus source (PRE-synaptic) coordinate value
        (i.e., POST-synaptic neuron below (delta < 0) or above (delta > 0) PRE-synaptic neuron assuming
         axis values increasing from lower to upper layers)
        """
        bip_mat = np.sign(
            np.diff(
                np.meshgrid(src_pos[:, bip_coord], tgt_pos[:, bip_coord], indexing="ij"), axis=0
            )[0, :, :]
        )  # Bipolar distinction based on difference in specified coordinate
        return bip_mat

    def get_model_output(self, src_pos, tgt_pos):  # pylint: disable=arguments-differ
        """Return (bipolar distance-dependent) connection probabilities <#src x #tgt> for all combinations of source/target neuron positions <#src/#tgt x #dim>."""
        #         log.log_assert(
        #             src_pos.shape[1] == tgt_pos.shape[1],
        #             "Dimension mismatch of source/target neuron positions!",
        #         )
        dist_mat = self.compute_dist_matrix(src_pos, tgt_pos)
        bip_mat = self.compute_bip_matrix(src_pos, tgt_pos, self.bip_coord)
        return self.get_conn_prob(dist_mat, bip_mat)

    def __str__(self):
        """Return model string describing the model."""
        model_str = f"{self.__class__.__name__}\n"
        model_str = (
            model_str
            + f"  p_conn(d, delta) = {self.prox_scale_N:.3f} * exp(-{self.prox_exp_N:.6f} * d^{self.prox_exp_pow_N:.3f}) + {self.dist_scale_N:.3f} * exp(-{self.dist_exp_N:.3f} * d) if delta < 0\n"
        )
        model_str = (
            model_str
            + f"                     {self.prox_scale_P:.3f} * exp(-{self.prox_exp_P:.6f} * d^{self.prox_exp_pow_P:.3f}) + {self.dist_scale_P:.3f} * exp(-{self.dist_exp_P:.3f} * d) if delta > 0\n"
        )
        model_str = model_str + "                     AVERAGE OF BOTH MODELS  if delta == 0\n"
        model_str = (
            model_str
            + f"  d...distance, delta...difference (tgt minus src) in coordinate {self.bip_coord}"
        )
        return model_str


class ConnProb4thOrderLinInterpnModel(AbstractModel):
    """4th order connection probability model (offset-dependent, linearly interpolated):

    - Returns (offset-dependent) connection probabilities for given source/target neuron positions
    """

    # Names of model inputs, parameters and data frames which are part of this model
    param_names = []
    param_defaults = {}
    data_names = ["p_conn_table"]
    input_names = ["src_pos", "tgt_pos"]

    def __init__(self, **kwargs):
        """Model initialization."""
        super().__init__(**kwargs)

        # Check parameters
        log.log_assert(
            len(self.p_conn_table.index.levels) == 3,
            "Data frame with 3 index levels (dx, dy, dz) required!",
        )
        log.log_assert(
            self.p_conn_table.shape[1] == 1,
            "Data frame with 1 column (conn. prob.) required!",
        )

        self._data_points = [
            list(lev_pos) for lev_pos in self.p_conn_table.index.levels
        ]  # Extract data offsets from multi-index
        self.p_data = self.p_conn_table.to_numpy().reshape(self.p_conn_table.index.levshape)
        self.data_dim_sel = (
            np.array(self.p_data.shape) > 1
        )  # Select data dimensions to be interpolated (removing dimensions with only a single data point from interpolation)

    @property
    def data_points(self):
        """Return data offsets."""
        return self._data_points

    def get_prob_data(self):
        """Return connection probability data."""
        return self.p_data

    def get_conn_prob(self, dx, dy, dz):
        """Return (offset-dependent) connection probability, linearly interpolating between data points (except dimensions with a single data point)."""
        data_sel = [val for idx, val in enumerate(self._data_points) if self.data_dim_sel[idx]]
        inp_sel = [
            val
            for idx, val in enumerate([np.array(dx), np.array(dy), np.array(dz)])
            if self.data_dim_sel[idx]
        ]

        # p_conn = np.minimum(np.maximum(interpn(data_sel, np.squeeze(self.p_data), np.array(inp_sel).T, method="linear", bounds_error=False, fill_value=None), 0.0), 1.0).T
        # [BUG]: Extrapolation artifacts under some circumstances [ACCS-37]

        # [FIX]: Don't use extrapolation, but set to zero instead
        #      + check that probability values at borders are sufficiently small, so that no (big) jumps at borders
        #        (otherwise, sampling range should be increased)
        p_conn = np.minimum(
            np.maximum(
                interpn(
                    data_sel,
                    np.squeeze(self.p_data),
                    np.array(inp_sel).T,
                    method="linear",
                    bounds_error=False,
                    fill_value=0.0,
                ),
                0.0,
            ),
            1.0,
        ).T

        # Check max. probability at boders (at dimensions that are actually interpolated)
        p_border = 0.0
        for dim in np.where(self.data_dim_sel)[0]:
            p_border = np.maximum(
                p_border, np.max(np.max(self.p_data, dim)[[0, -1]])
            )  # Take first/last element per dimension
        if p_border > P_TH_ABS or p_border / np.max(self.p_data) > P_TH_REL:
            log.warning(
                f"Probability at border should be close to zero (p_abs={p_border:.2f} (th_abs={P_TH_ABS:.2f}); p_rel={p_border / np.max(self.p_data):.2f} (th_rel={P_TH_REL:.2f})). Consider smoothing and/or increasing max. sampling range!"
            )

        return p_conn

    @staticmethod
    def compute_offset_matrices(src_pos, tgt_pos):
        """Computes dx/dy/dz offset matrices between pairs of neurons (tgt/POST minus src/PRE position)."""
        dx_mat = np.diff(np.meshgrid(src_pos[:, 0], tgt_pos[:, 0], indexing="ij"), axis=0)[
            0, :, :
        ]  # Relative difference in x coordinate
        dy_mat = np.diff(np.meshgrid(src_pos[:, 1], tgt_pos[:, 1], indexing="ij"), axis=0)[
            0, :, :
        ]  # Relative difference in y coordinate
        dz_mat = np.diff(np.meshgrid(src_pos[:, 2], tgt_pos[:, 2], indexing="ij"), axis=0)[
            0, :, :
        ]  # Relative difference in z coordinate
        return dx_mat, dy_mat, dz_mat

    def get_model_output(self, src_pos, tgt_pos):  # pylint: disable=arguments-differ
        """Return (offset-dependent) connection probabilities <#src x #tgt> for all combinations of source/target neuron positions <#src/#tgt x #dim>."""
        #         log.log_assert(
        #             src_pos.shape[1] == tgt_pos.shape[1],
        #             "Dimension mismatch of source/target neuron positions!",
        #         )
        #         log.log_assert(
        #             src_pos.shape[1] == 3, "Wrong number of input dimensions (3 required)!"
        #         )
        dx_mat, dy_mat, dz_mat = self.compute_offset_matrices(src_pos, tgt_pos)
        return self.get_conn_prob(dx_mat, dy_mat, dz_mat)

    def __str__(self):
        """Return model string describing the model."""
        inp_names = np.array(["dx", "dy", "dz"])
        inp_str = ", ".join(inp_names[self.data_dim_sel])
        range_str = ", ".join(
            [
                f"{inp_names[i]}({len(self.p_conn_table.index.levels[i])}): {np.min(self.p_conn_table.index.levels[i]):.2f}..{np.max(self.p_conn_table.index.levels[i]):.2f}"
                for i in range(self.p_conn_table.index.nlevels)
                if self.data_dim_sel[i]
            ]
        )
        model_str = f"{self.__class__.__name__}\n"
        model_str = (
            model_str
            + f"  p_conn({inp_str}) = LINEAR INTERPOLATION FROM DATA TABLE ({self.p_conn_table.shape[0]} entries; {range_str})\n"
        )
        model_str = model_str + "  dx/dy/dz...position offset (tgt minus src) in x/y/z dimension"
        return model_str


class ConnProb4thOrderLinInterpnReducedModel(AbstractModel):
    """Reduced 4th order connection probability model (offset-dependent in radial/axial direction, linearly interpolated):

    - Returns (offset-dependent) connection probabilities for given source/target neuron positions
    """

    # Names of model inputs, parameters and data frames which are part of this model
    param_names = ["axial_coord"]
    param_defaults = {}
    data_names = ["p_conn_table"]
    input_names = ["src_pos", "tgt_pos"]

    def __init__(self, **kwargs):
        """Model initialization."""
        super().__init__(**kwargs)

        # Check parameters
        log.log_assert(
            isinstance(self.axial_coord, int) and 0 <= self.axial_coord <= 2,
            'Axial coordinate "axial_coord" out of range!',
        )
        log.log_assert(
            len(self.p_conn_table.index.levels) == 2,
            "Data frame with 2 index levels (dr, dz) required!",
        )
        log.log_assert(
            self.p_conn_table.shape[1] == 1,
            "Data frame with 1 column (conn. prob.) required!",
        )

        self._data_points = [
            list(lev_pos) for lev_pos in self.p_conn_table.index.levels
        ]  # Extract data offsets from multi-index
        self.p_data = self.p_conn_table.to_numpy().reshape(self.p_conn_table.index.levshape)
        self.data_dim_sel = (
            np.array(self.p_data.shape) > 1
        )  # Select data dimensions to be interpolated (removing dimensions with only a single data point from interpolation)

        # Mirror around dr == 0, so that smooth interpolation for dr towards 0
        if self.data_dim_sel[0]:
            self._data_points[0] = [-d for d in self._data_points[0][::-1]] + self._data_points[0]
            self.p_data = np.vstack([self.p_data[::-1, :], self.p_data])

    @property
    def data_points(self):
        """Return data offsets."""
        return self._data_points

    def get_prob_data(self):
        """Return connection probability data."""
        return self.p_data

    def get_conn_prob(self, dr, dz):
        """Return (offset-dependent) connection probability, linearly interpolating between data points (except dimensions with a single data point)."""
        data_sel = [val for idx, val in enumerate(self._data_points) if self.data_dim_sel[idx]]
        inp_sel = [
            val for idx, val in enumerate([np.array(dr), np.array(dz)]) if self.data_dim_sel[idx]
        ]

        # p_conn = np.minimum(np.maximum(interpn(data_sel, np.squeeze(self.p_data), np.array(inp_sel).T, method="linear", bounds_error=False, fill_value=None), 0.0), 1.0).T
        # [BUG]: Extrapolation artifacts under some circumstances [ACCS-37]

        # [FIX]: Don't use extrapolation, but set to zero instead
        #      + check that probability values at borders are sufficiently small, so that no (big) jumps at borders
        #        (otherwise, sampling range should be increased)
        p_conn = np.minimum(
            np.maximum(
                interpn(
                    data_sel,
                    np.squeeze(self.p_data),
                    np.array(inp_sel).T,
                    method="linear",
                    bounds_error=False,
                    fill_value=0.0,
                ),
                0.0,
            ),
            1.0,
        ).T

        # Check max. probability at boders (at dimensions that are actually interpolated)
        p_border = 0.0
        for dim in np.where(self.data_dim_sel)[0]:
            p_border = np.maximum(
                p_border, np.max(np.max(self.p_data, dim)[[0, -1]])
            )  # Take first/last element per dimension
        if p_border > P_TH_ABS or p_border / np.max(self.p_data) > P_TH_REL:
            log.warning(
                f"Probability at border should be close to zero (p_abs={p_border:.2f} (th_abs={P_TH_ABS:.2f}); p_rel={p_border / np.max(self.p_data):.2f} (th_rel={P_TH_REL:.2f})). Consider smoothing and/or increasing max. sampling range!"
            )

        return p_conn

    @staticmethod
    def compute_offset_matrices(src_pos, tgt_pos, axial_coord):
        """Computes radial/axial offset matrices between pairs of neurons (tgt/POST minus src/PRE position)."""
        d_matrices = [
            np.diff(np.meshgrid(src_pos[:, i], tgt_pos[:, i], indexing="ij"), axis=0)[0, :, :]
            for i in range(3)
        ]  # Relative differences in x/y/z coordinates

        radial_coords = list(set(range(3)) - {axial_coord})  # Radial coordinates
        dr_mat = np.sqrt(
            d_matrices[radial_coords[0]] ** 2 + d_matrices[radial_coords[1]] ** 2
        )  # Relative offset in radial plane (Euclidean distance)
        dz_mat = d_matrices[axial_coord]  # Axial offset

        return dr_mat, dz_mat

    def get_model_output(self, src_pos, tgt_pos):  # pylint: disable=arguments-differ
        """Return (offset-dependent) connection probabilities <#src x #tgt> for all combinations of source/target neuron positions <#src/#tgt x #dim>."""
        #         log.log_assert(
        #             src_pos.shape[1] == tgt_pos.shape[1],
        #             "Dimension mismatch of source/target neuron positions!",
        #         )
        #         log.log_assert(
        #             src_pos.shape[1] == 3, "Wrong number of input dimensions (3 required)!"
        #         )
        dr_mat, dz_mat = self.compute_offset_matrices(src_pos, tgt_pos, self.axial_coord)
        return self.get_conn_prob(dr_mat, dz_mat)

    def __str__(self):
        """Return model string describing the model."""
        inp_names = np.array(["dr", "dz"])
        inp_str = ", ".join(inp_names[self.data_dim_sel])
        range_str = ", ".join(
            [
                f"{inp_names[i]}({len(self.p_conn_table.index.levels[i])}): {np.min(self.p_conn_table.index.levels[i]):.2f}..{np.max(self.p_conn_table.index.levels[i]):.2f}"
                for i in range(self.p_conn_table.index.nlevels)
                if self.data_dim_sel[i]
            ]
        )
        model_str = f"{self.__class__.__name__}\n"
        model_str = (
            model_str
            + f"  p_conn({inp_str}) = LINEAR INTERPOLATION FROM DATA TABLE ({self.p_conn_table.shape[0]} entries; {range_str})\n"
        )
        model_str = (
            model_str
            + f"  dr/dz...radial/axial position offset (tgt minus src), with axial coordinate {self.axial_coord}"
        )
        return model_str


class ConnProb5thOrderLinInterpnModel(AbstractModel):
    """5th order connection probability model (position- & offset-dependent, linearly interpolated):

    - Returns (position- & offset-dependent) connection probabilities for given source/target neuron positions
    """

    # Names of model inputs, parameters and data frames which are part of this model
    param_names = []
    param_defaults = {}
    data_names = ["p_conn_table"]
    input_names = ["src_pos", "tgt_pos"]

    def __init__(self, **kwargs):
        """Model initialization."""
        super().__init__(**kwargs)

        # Check parameters
        log.log_assert(
            len(self.p_conn_table.index.levels) == 6,
            "Data frame with 6 index levels (x, y, z, dx, dy, dz) required!",
        )
        log.log_assert(
            self.p_conn_table.shape[1] == 1,
            "Data frame with 1 column (conn. prob.) required!",
        )

        self._data_points = [
            list(lev_pos) for lev_pos in self.p_conn_table.index.levels
        ]  # Extract data positions & offsets from multi-index
        self.p_data = self.p_conn_table.to_numpy().reshape(self.p_conn_table.index.levshape)
        self.data_dim_sel = (
            np.array(self.p_data.shape) > 1
        )  # Select data dimensions to be interpolated (removing dimensions with only a single data point from interpolation)

    @property
    def data_points(self):
        """Return data offsets."""
        return self._data_points

    def get_prob_data(self):
        """Return connection probability data."""
        return self.p_data

    def get_conn_prob(self, x, y, z, dx, dy, dz):
        """Return (position- & offset-dependent) connection probability, linearly interpolating between data points (except dimensions with a single data point)."""
        data_sel = [val for idx, val in enumerate(self._data_points) if self.data_dim_sel[idx]]
        inp_sel = [
            val
            for idx, val in enumerate(
                [np.array(x), np.array(y), np.array(z), np.array(dx), np.array(dy), np.array(dz)]
            )
            if self.data_dim_sel[idx]
        ]

        # p_conn = np.minimum(np.maximum(interpn(data_sel, np.squeeze(self.p_data), np.array(inp_sel).T, method="linear", bounds_error=False, fill_value=None), 0.0), 1.0).T
        # [BUG]: Extrapolation artifacts under some circumstances [ACCS-37]

        # [FIX]: Don't use extrapolation, but set to zero instead
        #      + check that probability values at borders are sufficiently small, so that no (big) jumps at borders
        #        (otherwise, sampling range should be increased)
        p_conn = np.minimum(
            np.maximum(
                interpn(
                    data_sel,
                    np.squeeze(self.p_data),
                    np.array(inp_sel).T,
                    method="linear",
                    bounds_error=False,
                    fill_value=0.0,
                ),
                0.0,
            ),
            1.0,
        ).T

        # Check max. probability at boders (at dimensions that are actually interpolated)
        p_border = 0.0
        for dim in np.where(self.data_dim_sel)[0]:
            p_border = np.maximum(
                p_border, np.max(np.max(self.p_data, dim)[[0, -1]])
            )  # Take first/last element per dimension
        if p_border > P_TH_ABS or p_border / np.max(self.p_data) > P_TH_REL:
            log.warning(
                f"Probability at border should be close to zero (p_abs={p_border:.2f} (th_abs={P_TH_ABS:.2f}); p_rel={p_border / np.max(self.p_data):.2f} (th_rel={P_TH_REL:.2f})). Consider smoothing and/or increasing max. sampling range!"
            )

        return p_conn

    @staticmethod
    def compute_position_matrices(src_pos, tgt_pos):
        """Computes x/y/z position matrices of src/PRE neurons (src/PRE neuron positions repeated over tgt/POST neuron number)."""
        x_mat, y_mat, z_mat = [
            np.tile(src_pos[:, i : i + 1], [1, tgt_pos.shape[0]]) for i in range(src_pos.shape[1])
        ]
        return x_mat, y_mat, z_mat

    @staticmethod
    def compute_offset_matrices(src_pos, tgt_pos):
        """Computes dx/dy/dz offset matrices between pairs of neurons (tgt/POST minus src/PRE position)."""
        dx_mat = np.diff(np.meshgrid(src_pos[:, 0], tgt_pos[:, 0], indexing="ij"), axis=0)[
            0, :, :
        ]  # Relative difference in x coordinate
        dy_mat = np.diff(np.meshgrid(src_pos[:, 1], tgt_pos[:, 1], indexing="ij"), axis=0)[
            0, :, :
        ]  # Relative difference in y coordinate
        dz_mat = np.diff(np.meshgrid(src_pos[:, 2], tgt_pos[:, 2], indexing="ij"), axis=0)[
            0, :, :
        ]  # Relative difference in z coordinate
        return dx_mat, dy_mat, dz_mat

    def get_model_output(self, src_pos, tgt_pos):  # pylint: disable=arguments-differ
        """Return (position- & offset-dependent) connection probabilities <#src x #tgt> for all combinations of source/target neuron positions <#src/#tgt x #dim>."""
        #         log.log_assert(
        #             src_pos.shape[1] == tgt_pos.shape[1],
        #             "Dimension mismatch of source/target neuron positions!",
        #         )
        #         log.log_assert(
        #             src_pos.shape[1] == 3, "Wrong number of input dimensions (3 required)!"
        #         )
        x_mat, y_mat, z_mat = self.compute_position_matrices(src_pos, tgt_pos)
        dx_mat, dy_mat, dz_mat = self.compute_offset_matrices(src_pos, tgt_pos)
        return self.get_conn_prob(x_mat, y_mat, z_mat, dx_mat, dy_mat, dz_mat)

    def __str__(self):
        """Return model string describing the model."""
        inp_names = np.array(["x", "y", "z", "dx", "dy", "dz"])
        inp_str = ", ".join(inp_names[self.data_dim_sel])
        range_str = ", ".join(
            [
                f"{inp_names[i]}({len(self.p_conn_table.index.levels[i])}): {np.min(self.p_conn_table.index.levels[i]):.2f}..{np.max(self.p_conn_table.index.levels[i]):.2f}"
                for i in range(self.p_conn_table.index.nlevels)
                if self.data_dim_sel[i]
            ]
        )
        model_str = f"{self.__class__.__name__}\n"
        model_str = (
            model_str
            + f"  p_conn({inp_str}) = LINEAR INTERPOLATION FROM DATA TABLE ({self.p_conn_table.shape[0]} entries; {range_str})\n"
        )
        model_str = (
            model_str
            + "  x/y/z...src position, dx/dy/dz...position offset (tgt minus src) in x/y/z dimension"
        )
        return model_str


class ConnProb5thOrderLinInterpnReducedModel(AbstractModel):
    """Reduced 5th order connection probability model (position-dependent in axial direction & offset-dependent in radial/axial direction, linearly interpolated):

    - Returns (position- & offset-dependent) connection probabilities for given source/target neuron positions
    """

    # Names of model inputs, parameters and data frames which are part of this model
    param_names = ["axial_coord"]
    param_defaults = {}
    data_names = ["p_conn_table"]
    input_names = ["src_pos", "tgt_pos"]

    def __init__(self, **kwargs):
        """Model initialization."""
        super().__init__(**kwargs)

        # Check parameters
        log.log_assert(
            isinstance(self.axial_coord, int) and 0 <= self.axial_coord <= 2,
            'Axial coordinate "axial_coord" out of range!',
        )
        log.log_assert(
            len(self.p_conn_table.index.levels) == 3,
            "Data frame with 3 index levels (z, dr, dz) required!",
        )
        log.log_assert(
            self.p_conn_table.shape[1] == 1,
            "Data frame with 1 column (conn. prob.) required!",
        )

        self._data_points = [
            list(lev_pos) for lev_pos in self.p_conn_table.index.levels
        ]  # Extract data positions & offsets from multi-index
        self.p_data = self.p_conn_table.to_numpy().reshape(self.p_conn_table.index.levshape)
        self.data_dim_sel = (
            np.array(self.p_data.shape) > 1
        )  # Select data dimensions to be interpolated (removing dimensions with only a single data point from interpolation)

        # Mirror around dr == 0, so that smooth interpolation for dr towards 0
        if self.data_dim_sel[1]:
            self._data_points[1] = [-d for d in self._data_points[1][::-1]] + self._data_points[1]
            self.p_data = np.hstack([self.p_data[:, ::-1, :], self.p_data])

    @property
    def data_points(self):
        """Return data offsets."""
        return self._data_points

    def get_prob_data(self):
        """Return connection probability data."""
        return self.p_data

    def get_conn_prob(self, z, dr, dz):
        """Return (position- & offset-dependent) connection probability, linearly interpolating between data points (except dimensions with a single data point)."""
        data_sel = [val for idx, val in enumerate(self._data_points) if self.data_dim_sel[idx]]
        inp_sel = [
            val
            for idx, val in enumerate([np.array(z), np.array(dr), np.array(dz)])
            if self.data_dim_sel[idx]
        ]

        # p_conn = np.minimum(np.maximum(interpn(data_sel, np.squeeze(self.p_data), np.array(inp_sel).T, method="linear", bounds_error=False, fill_value=None), 0.0), 1.0).T
        # [BUG]: Extrapolation artifacts under some circumstances [ACCS-37]

        # [FIX]: Don't use extrapolation, but set to zero instead
        #      + check that probability values at borders are sufficiently small, so that no (big) jumps at borders
        #        (otherwise, sampling range should be increased)
        p_conn = np.minimum(
            np.maximum(
                interpn(
                    data_sel,
                    np.squeeze(self.p_data),
                    np.array(inp_sel).T,
                    method="linear",
                    bounds_error=False,
                    fill_value=0.0,
                ),
                0.0,
            ),
            1.0,
        ).T

        # Check max. probability at boders (at dimensions that are actually interpolated)
        p_border = 0.0
        for dim in np.where(self.data_dim_sel)[0]:
            p_border = np.maximum(
                p_border, np.max(np.max(self.p_data, dim)[[0, -1]])
            )  # Take first/last element per dimension
        if p_border > P_TH_ABS or p_border / np.max(self.p_data) > P_TH_REL:
            log.warning(
                f"Probability at border should be close to zero (p_abs={p_border:.2f} (th_abs={P_TH_ABS:.2f}); p_rel={p_border / np.max(self.p_data):.2f} (th_rel={P_TH_REL:.2f})). Consider smoothing and/or increasing max. sampling range!"
            )

        return p_conn

    @staticmethod
    def compute_position_matrix(src_pos, tgt_pos, axial_coord):
        """Computes axial position matrix of src/PRE neurons (src/PRE neuron positions repeated over tgt/POST neuron number)."""
        z_mat = np.tile(src_pos[:, [axial_coord]], [1, tgt_pos.shape[0]])  # Axial position
        return z_mat

    @staticmethod
    def compute_offset_matrices(src_pos, tgt_pos, axial_coord):
        """Computes radial/axial offset matrices between pairs of neurons (tgt/POST minus src/PRE position)."""
        d_matrices = [
            np.diff(np.meshgrid(src_pos[:, i], tgt_pos[:, i], indexing="ij"), axis=0)[0, :, :]
            for i in range(3)
        ]  # Relative differences in x/y/z coordinates

        radial_coords = list(set(range(3)) - {axial_coord})  # Radial coordinates
        dr_mat = np.sqrt(
            d_matrices[radial_coords[0]] ** 2 + d_matrices[radial_coords[1]] ** 2
        )  # Relative offset in radial plane (Euclidean distance)
        dz_mat = d_matrices[axial_coord]  # Axial offset

        return dr_mat, dz_mat

    def get_model_output(self, src_pos, tgt_pos):  # pylint: disable=arguments-differ
        """Return (position- & offset-dependent) connection probabilities <#src x #tgt> for all combinations of source/target neuron positions <#src/#tgt x #dim>."""
        #         log.log_assert(
        #             src_pos.shape[1] == tgt_pos.shape[1],
        #             "Dimension mismatch of source/target neuron positions!",
        #         )
        #         log.log_assert(
        #             src_pos.shape[1] == 3, "Wrong number of input dimensions (3 required)!"
        #         )
        z_mat = self.compute_position_matrix(src_pos, tgt_pos, self.axial_coord)
        dr_mat, dz_mat = self.compute_offset_matrices(src_pos, tgt_pos, self.axial_coord)
        return self.get_conn_prob(z_mat, dr_mat, dz_mat)

    def __str__(self):
        """Return model string describing the model."""
        inp_names = np.array(["z", "dr", "dz"])
        inp_str = ", ".join(inp_names[self.data_dim_sel])
        range_str = ", ".join(
            [
                f"{inp_names[i]}({len(self.p_conn_table.index.levels[i])}): {np.min(self.p_conn_table.index.levels[i]):.2f}..{np.max(self.p_conn_table.index.levels[i]):.2f}"
                for i in range(self.p_conn_table.index.nlevels)
                if self.data_dim_sel[i]
            ]
        )
        model_str = f"{self.__class__.__name__}\n"
        model_str = (
            model_str
            + f"  p_conn({inp_str}) = LINEAR INTERPOLATION FROM DATA TABLE ({self.p_conn_table.shape[0]} entries; {range_str})\n"
        )
        model_str = (
            model_str
            + f"  z...axial src position, dr/dz...radial/axial position offset (tgt minus src), with axial coordinate {self.axial_coord}"
        )
        return model_str


class LookupTableModel(AbstractModel):
    """Generic model to access any (sparse) information based on source and target neuron IDs.

    This model can be used for adjacency matrices (bool), deterministic connection
    probabilities (float), specific #syn/conn (int), etc.
    - Data are internally stored as sparse matrix in CSC format
      (initialized from data frame with 'row_ind', 'col_ind', and 'value' columns)
    - Size of that matrix matches selected src/dest neuron selections
      (initialized from data frames with 'src_node_ids' and 'tgt_node_ids')
    - Returns (deterministic) value (always float) for given source/target neuron IDs
      (or 0.0 if no value is stored)
    """

    # Names of model inputs, parameters and data frames which are part of this model
    param_names = []
    param_defaults = {}
    data_names = ["src_nodes_table", "tgt_nodes_table", "lookup_table"]
    input_names = ["src_nid", "tgt_nid"]

    def __init__(self, **kwargs):
        """Model initialization."""
        super().__init__(**kwargs)

        # Check parameters
        log.log_assert(
            self.src_nodes_table.shape[1] == 1,
            "Data frame with 1 column (src_node_ids) required!",
        )
        self.src_node_ids = self.src_nodes_table.to_numpy().flatten()
        self.src_idx_table = pd.Series(
            self.src_nodes_table.index, index=self.src_node_ids, name="src_index"
        )

        log.log_assert(
            self.tgt_nodes_table.shape[1] == 1,
            "Data frame with 1 column (tgt_node_ids) required!",
        )
        self.tgt_node_ids = self.tgt_nodes_table.to_numpy().flatten()
        self.tgt_idx_table = pd.Series(
            self.tgt_nodes_table.index, index=self.tgt_node_ids, name="tgt_index"
        )

        log.log_assert(
            self.lookup_table.shape[1] == 3,
            "Data frame with 3 columns (row_ind, col_ind, value) required!",
        )
        self.lut_mat = csc_matrix(
            (
                self.lookup_table["value"],
                (self.lookup_table["row_ind"], self.lookup_table["col_ind"]),
            ),
            shape=(len(self.src_nodes_table), len(self.tgt_nodes_table)),
        )

    def get_lookup_table(self):
        """Return underlying (sparse) LUT matrix."""
        return self.lut_mat

    def get_src_nids(self):
        """Return source node IDs stored in this model."""
        return self.src_node_ids

    def get_tgt_nids(self):
        """Return target node IDs stored in this model."""
        return self.tgt_node_ids

    def get_model_output(self, src_nid, tgt_nid):  # pylint: disable=arguments-differ
        """Return LUT values of size <#src x #tgt> as given by underlying LUT matrix."""
        src_sel = self.src_idx_table.loc[src_nid].values
        tgt_sel = self.tgt_idx_table.loc[tgt_nid].values
        mat_sel = self.lut_mat[:, tgt_sel][src_sel, :].toarray()
        return mat_sel

    def __str__(self):
        """Return model string describing the model."""
        model_str = f"{self.__class__.__name__}"
        model_str = model_str + "\n  " + self.lut_mat.__repr__()
        if len(self.lut_mat.data) > 0:
            range_str = f"{self.lut_mat.data.min()}"
            if self.lut_mat.data.max() > self.lut_mat.data.min():
                range_str = range_str + f"..{self.lut_mat.data.max()}"
        else:
            range_str = "None"
        model_str = model_str + "\n  Value range: " + range_str
        model_str = model_str + f" (dtype: {self.lut_mat.dtype})"
        return model_str

    @staticmethod
    def init_from_sparse_matrix(matrix, src_node_ids, tgt_node_ids):
        """Model initialization from sparse matrix."""
        log.log_assert(sps.issparse(matrix), "Matrix must be in sparse format!")
        log.log_assert(matrix.shape[0] == len(src_node_ids), "Source node IDs mismatch!")
        log.log_assert(matrix.shape[1] == len(tgt_node_ids), "Target node IDs mismatch!")

        src_nodes_table = pd.DataFrame(src_node_ids, columns=["src_node_ids"])
        tgt_nodes_table = pd.DataFrame(tgt_node_ids, columns=["tgt_node_ids"])
        matrix = matrix.tocoo()  # Convert to COO, for easy access to row/col and data!!
        lookup_table = pd.DataFrame(
            {"row_ind": matrix.row, "col_ind": matrix.col, "value": matrix.data}
        )

        return LookupTableModel(
            src_nodes_table=src_nodes_table,
            tgt_nodes_table=tgt_nodes_table,
            lookup_table=lookup_table,
        )


class ConnProbAdjModel(AbstractModel):
    """Deterministic connection probability model, defined by an adjacency matrix (internally stored as LookupTableModel):

    - Adjacency matrix represented as boolean matrix in sparse CSC format
      (initialized from data frame with 'row_ind' and 'col_ind' columns)
    - Size of adjacency matrix matches selected src/dest neuron selections
      (initialized from data frames with 'src_node_ids' and 'tgt_node_ids')
    - Returns deterministic connection probability (0.0 or 1.0) for given source/target neuron IDs
    - Can optionally store inverted connectivity (i.e., True...no connection, False...connection)
    """

    # Names of model inputs, parameters and data frames which are part of this model
    param_names = ["inverted"]
    param_defaults = {"inverted": False}
    data_names = ["src_nodes_table", "tgt_nodes_table", "adj_table"]
    input_names = ["src_nid", "tgt_nid"]

    def __init__(self, **kwargs):
        """Model initialization."""
        super().__init__(**kwargs)

        # Check parameters
        log.log_assert(
            isinstance(self.inverted, bool),
            "Inverted flag must be boolean!",
        )

        # Init internal LUT
        lookup_table = self.adj_table.copy()
        lookup_table["value"] = [True] * lookup_table.shape[0]
        self.lut = LookupTableModel(
            src_nodes_table=self.src_nodes_table,
            tgt_nodes_table=self.tgt_nodes_table,
            lookup_table=lookup_table,
        )

    def get_adj_matrix(self):
        """Return adjacency matrix."""
        return self.lut.get_lookup_table()

    def get_src_nids(self):
        """Return source node IDs stored in this model."""
        return self.lut.src_node_ids

    def get_tgt_nids(self):
        """Return target node IDs stored in this model."""
        return self.lut.tgt_node_ids

    def is_inverted(self):
        """Return if connectivity is stored inverted."""
        return self.inverted

    def get_model_output(self, src_nid, tgt_nid):  # pylint: disable=arguments-differ
        """Return deterministic connection probabilities (0.0 or 1.0) of size <#src x #tgt> as given by adjacency matrix."""
        adj_sel = self.lut.get_model_output(src_nid, tgt_nid)
        if self.inverted:
            return np.logical_not(adj_sel).astype(float)
        else:
            return adj_sel.astype(float)

    def __str__(self):
        """Return model string describing the model."""
        if self.inverted:
            inv_str = "Inverted "
        else:
            inv_str = ""
        model_str = f"{inv_str}{self.__class__.__name__}\n"
        model_str = model_str + "  " + self.lut.get_lookup_table().__repr__()
        return model_str


class PropsTableModel(AbstractModel):
    """Generic model to store any/multiple (edge) properties based on source and target neuron IDs.

    e.g., this model can be used for providing synapse positions, etc.
    - Data are initialized and internally stored as data frame with properties as columns
    - Source/target node IDs must be provided as @source_node/@target_node columns
    """

    # Names of model inputs, parameters and data frames which are part of this model
    param_names = []
    param_defaults = {}
    data_names = ["props_table"]
    input_names = ["src_nid", "tgt_nid", "prop_names", "num_sel"]
    # Notes: "prop_names" is optional; if not provided, all properties will be returned
    #        "num_sel" is optional; if not provided, all entries will be returned; otherwise, only the first num_sel

    def __init__(self, **kwargs):
        """Model initialization."""
        super().__init__(**kwargs)

        # Check parameters
        log.log_assert("@source_node" in self.props_table.columns, "@source_node column required!")
        log.log_assert("@target_node" in self.props_table.columns, "@target_node column required!")
        self.property_names = np.setdiff1d(
            self.props_table.columns, ["@source_node", "@target_node"]
        ).tolist()
        log.log_assert(len(self.property_names) > 0, "Properties missing!")
        log.log_assert(self.props_table.size > 0, "Properties table empty!")

    def get_property_names(self):
        """Return property names."""
        return self.property_names

    def get_src_nids(self):
        """Return (unique) source node IDs stored in this model."""
        return np.unique(self.props_table["@source_node"])

    def get_tgt_nids(self):
        """Return (unique) target node IDs stored in this model."""
        return np.unique(self.props_table["@target_node"])

    def get_src_tgt_counts(self):
        """Return (unique) source/target node ID pairs and counts stored in this model."""
        return np.unique(
            self.props_table[["@source_node", "@target_node"]], axis=0, return_counts=True
        )

    def get_model_output(
        self, src_nid, tgt_nid, prop_names=None, num_sel=None
    ):  # pylint: disable=arguments-differ
        """Return property values of given source/target node IDs."""
        src_sel = np.isin(self.props_table["@source_node"], src_nid)
        tgt_sel = np.isin(self.props_table["@target_node"], tgt_nid)
        if prop_names is None:
            prop_names = self.property_names
        tab = self.props_table[prop_names].loc[np.logical_and(src_sel, tgt_sel)]
        if num_sel is None:
            return tab
        else:
            log.log_assert(
                0 <= num_sel <= tab.shape[0], "Selected number of elements out of range!"
            )
            return tab[:num_sel]

    def apply(self, **kwargs):
        """Apply the model. Overwrite to set default."""
        if "prop_names" not in kwargs:
            kwargs["prop_names"] = None
        if "num_sel" not in kwargs:
            kwargs["num_sel"] = None
        return super().apply(**kwargs)

    def __str__(self):
        """Return model string describing the model."""
        model_str = f"{self.__class__.__name__}"
        model_str = model_str + f"\n  Properties: {self.property_names}"
        model_str = model_str + f"\n  Entries: {self.props_table.shape[0]}"
        return model_str



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

import os

import numpy as np
from numpy.testing import assert_array_equal
from bluepysnap import Circuit
from voxcell import VoxelData

from utils import TEST_DATA_DIR
import connectome_manipulator.access_functions as test_module


def test_get_node_ids():
    circuit = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    nodes = circuit.nodes[circuit.nodes.population_names[0]]

    # Check direct node set access
    for node_set in list(circuit.node_sets.content.keys()):
        ids = test_module.get_node_ids(nodes, node_set)
        assert_array_equal(ids, nodes.ids(node_set))

    # Check node access based on properties (layer)
    layers = list(np.unique(nodes.get(properties="layer")))
    for lay in layers:
        ids = test_module.get_node_ids(nodes, {"layer": lay})
        assert_array_equal(ids, nodes.ids({"layer": lay}))

    # Check combination of node set and properties (layer)
    for node_set in list(circuit.node_sets.content.keys()):
        for lay in layers:
            ids = test_module.get_node_ids(nodes, {"node_set": node_set, "layer": lay})
            ref_ids = np.intersect1d(nodes.ids(node_set), nodes.ids({"layer": lay}))
            assert_array_equal(ids, ref_ids)


def test_get_edges_population():
    circuit = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    popul_names = circuit.edges.population_names

    # Check selecting single (default) population
    edges = test_module.get_edges_population(circuit)
    assert edges is circuit.edges[popul_names[0]]

    # Check returned population name
    edges, pname = test_module.get_edges_population(circuit, return_popul_name=True)
    assert edges is circuit.edges[popul_names[0]]
    assert pname == popul_names[0]

    # Check selecting population by name (in case of single population)
    edges = test_module.get_edges_population(circuit, popul_names[0])
    assert edges is circuit.edges[popul_names[0]]

    # Check selecting population by name (in case of multiple populations)
    # (Not implemented: Would require test circuit with multiple populations)


def test_get_node_positions():
    circuit = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    nodes = circuit.nodes[circuit.nodes.population_names[0]]
    np.random.seed(0)
    node_ids = np.random.permutation(nodes.ids())
    ref_pos = nodes.positions(node_ids).to_numpy()

    # Case 1: No voxel map provided
    res = test_module.get_node_positions(nodes, node_ids, vox_map=None)
    assert_array_equal(ref_pos, res[0])  # raw_pos
    assert_array_equal(ref_pos, res[1])  # pos

    # Case 2: Voxel map provided
    vox_map = VoxelData.load_nrrd(os.path.join(TEST_DATA_DIR, "xy_map_lin.nrrd"))
    res = test_module.get_node_positions(nodes, node_ids, vox_map=vox_map)
    assert_array_equal(ref_pos, res[0])  # raw_pos
    assert_array_equal(vox_map.lookup(ref_pos), res[1])  # pos



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

import os
import types

import numpy as np
from numpy.testing import assert_array_equal

from utils import setup_tempdir, TEST_DATA_DIR
import connectome_manipulator.model_building.model_building as test_module


# def test_get_model():
#     model = '(lambda a,b,c,d: (a,b,c,d))(test_arg_1, test_arg_2, test_kwarg_1, test_kwarg_2)'
#     model_inputs = ['test_arg_1', 'test_arg_2']
#     model_params = {'test_kwarg_1': 'value_1', 'test_kwarg_2': 'value_2'}

#     res = test_module.get_model(model, model_inputs, model_params)
#     assert isinstance(res, types.FunctionType)

#     args = ['test_value_1', 'test_value_2']
#     assert_array_equal(res(*args), args + list(model_params.values()))

#     # NOTE by herttuai on 08/10/2021:
#     # Is it intentional that kwargs are given as (kwargs) and not (**kwargs)?
#     kwargs = {'test_kwarg_1': 'test_value_3', 'test_kwarg_2': 'test_value_4'}
#     assert_array_equal(res(*args, kwargs), args + list(kwargs.values()))


def test_create_model_config_per_pathway():
    with setup_tempdir(__name__) as tempdir:
        model_config = {
            "model": {
                "name": "TestCreateModelConfigPerPathway",
                "fct": {"source": "test_model", "kwargs": {}},
            },
            "working_dir": os.path.join(tempdir, "working_dir"),
            "out_dir": os.path.join(tempdir, "out_dir"),
            "circuit_config": os.path.join(TEST_DATA_DIR, "circuit_sonata.json"),
        }

        res = test_module.create_model_config_per_pathway(model_config, "synapse_class")

        # Check for parameters that are same in each model
        assert all([c["circuit_config"] == model_config["circuit_config"] for c in res])

        for key in ["out_dir", "working_dir"]:
            assert len({c[key] for c in res}) == 1

        assert len({c["model"]["fct"]["source"] for c in res}) == 1

        # Check that models have different name
        assert len({c["model"]["name"] for c in res}) == len(res)

        # Check that kwargs are unique
        for i in range(len(res) - 1):
            for j in range(i + 1, len(res)):
                assert res[i]["model"]["fct"] != res[j]["model"]["fct"]



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""Main CLI entry point"""

from pathlib import Path
import logging
import json
import click

from . import utils, log, profiler
from .connectome_manipulation import connectome_manipulation
from .model_building import model_building
from .connectome_comparison import structural_comparator

logger = logging.getLogger(__name__)


@click.group(context_settings={"show_default": True})
@click.version_option()
@click.option("-v", "--verbose", count=True, default=0, help="-v for INFO, -vv for DEBUG")
def app(verbose):
    """Connectome manipulation tools."""
    level = min(verbose, 2)
    log.setup_logging(log_level=level)


@app.command()
@click.argument("config", required=True, type=Path)
@click.option("--output-dir", required=True, type=Path, help="Output directory.")
@click.option("--profile", required=False, is_flag=True, type=bool, help="Enable profiling.")
@click.option(
    "--resume",
    required=False,
    is_flag=True,
    type=bool,
    help="Resume from exisiting .parquet files.",
)
@click.option("--keep-parquet", required=False, is_flag=True, help="Keep temporary parquet files.")
@click.option(
    "--convert-to-sonata",
    required=False,
    is_flag=True,
    help="Convert parquet to sonata and generate circuit config",
)
@click.option(
    "--overwrite-edges", required=False, is_flag=True, help="Overwrite existing edges file"
)
@click.option(
    "--splits",
    required=False,
    default=0,
    type=int,
    help="Number of blocks, overwrites value in config file",
)
@click.option(
    "--target-payload",
    required=False,
    default=20_000_000_000,
    type=int,
    help="Number of gid-gid pairs to consider for one block.  Supersedes splits when a parquet based configuration is used",
)
@click.option(
    "--parallel",
    required=False,
    is_flag=True,
    help="Run using a parallel DASK job scheduler",
)
@click.option(
    "--parallel-arg",
    "-a",
    required=False,
    multiple=True,
    type=str,
    help="Overwrite the arguments for the Dask Client with key=value",
)
def manipulate_connectome(
    config,
    output_dir,
    profile,
    resume,
    keep_parquet,
    convert_to_sonata,
    overwrite_edges,
    splits,
    target_payload,
    parallel,
    parallel_arg,
):
    """Manipulate or build a circuit's connectome."""
    # until we start using verbosity with the logging refactoring
    # pylint: disable=unused-argument
    # Initialize logger
    logging_path = output_dir / "logs"
    log_file = log.create_log_file(logging_path, "connectome_manipulation")

    output_path = utils.create_dir(output_dir)
    profiler.ProfilerManager.set_enabled(profile)

    options = connectome_manipulation.Options(
        config_path=config,
        output_path=output_path,
        logging_path=logging_path,
        do_profiling=profile,
        do_resume=resume,
        keep_parquet=keep_parquet,
        convert_to_sonata=convert_to_sonata,
        overwrite_edges=overwrite_edges,
        splits=splits,
        target_payload=target_payload,
        parallel=parallel,
    )

    connectome_manipulation.main(options, log_file, executor_args=parallel_arg)
    profiler.ProfilerManager.show_stats()


@app.command()
@click.argument("config", required=True, type=Path)
@click.option(
    "--force-reextract",
    required=False,
    is_flag=True,
    type=bool,
    help="Force re-extraction of data, in case already existing.",
)
@click.option(
    "--force-rebuild",
    required=False,
    is_flag=True,
    type=bool,
    help="Force model re-building, in case already existing.",
)
@click.option(
    "--cv-folds",
    required=False,
    default=None,
    type=int,
    help="Optional number of cross-validation folds, overwrites value in config file",
)
def build_model(config, force_reextract, force_rebuild, cv_folds):
    """Extract and build models from existing connectomes."""
    with open(config, "r") as f:
        config_dict = json.load(f)

    model_building.main(
        config_dict,
        show_fig=False,
        force_recomp=[force_reextract, force_rebuild],
        cv_folds=cv_folds,
    )


@app.command()
@click.argument("config", required=True, type=Path)
@click.option(
    "--force-recomp-circ1",
    required=False,
    is_flag=True,
    type=bool,
    help="Force re-computation of 1st circuit's comparison data, in case already existing.",
)
@click.option(
    "--force-recomp-circ2",
    required=False,
    is_flag=True,
    type=bool,
    help="Force re-computation of 2nd circuit's comparison data, in case already existing.",
)
def compare_connectomes(config, force_recomp_circ1, force_recomp_circ2):
    """Compare connectome structure of two circuits."""
    with open(config, "r") as f:
        config_dict = json.load(f)

    structural_comparator.main(
        config_dict, show_fig=False, force_recomp=[force_recomp_circ1, force_recomp_circ2]
    )


if __name__ == "__main__":
    app()  # pylint: disable=no-value-for-parameter



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

import os

import numpy as np
import pandas as pd
import pytest

from bluepysnap.morph import MorphHelper
from bluepysnap import Circuit
from libsonata import Selection
import neurom as nm

from utils import TEST_DATA_DIR
from connectome_manipulator.model_building import model_types
from connectome_manipulator.connectome_manipulation.manipulation import Manipulation
from connectome_manipulator.connectome_manipulation.converters import EdgeWriter


# SONATA section type mapping (as in MorphIO): 1 = soma, 2 = axon, 3 = basal, 4 = apical
SEC_SOMA = 1
SEC_TYPE_MAP = {nm.AXON: 2, nm.BASAL_DENDRITE: 3, nm.APICAL_DENDRITE: 4}


@pytest.fixture
def manipulation():
    m = Manipulation.get("conn_wiring")
    return m


def test_apply(manipulation):
    c = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    edges = c.edges[c.edges.population_names[0]]
    nodes = [edges.source, edges.target]
    src_mtypes = nodes[0].property_values("mtype")
    tgt_mtypes = nodes[1].property_values("mtype")
    src_ids = nodes[0].ids()
    tgt_ids = nodes[1].ids()
    edges_table = edges.afferent_edges(tgt_ids, properties=edges.property_names)
    edges_table_empty = edges_table.loc[[]].copy()

    required_properties = [
        "@source_node",
        "@target_node",
        "afferent_center_x",
        "afferent_center_y",
        "afferent_center_z",
        "afferent_section_pos",
        "afferent_section_id",
        "afferent_section_type",
        "syn_type_id",
    ]

    tgt_morph = MorphHelper(
        nodes[1].config.get("morphologies_dir"),
        nodes[1],
        nodes[1].config.get("alternate_morphologies"),
    )
    get_tgt_morph = lambda node_id: tgt_morph.get(
        node_id, transform=True, extension="swc"
    )  # Access function (incl. transformation!), using specified format (swc/h5/...)

    n_syn_conn = 2
    nsynconn_model_file = os.path.join(
        TEST_DATA_DIR, f"model_config__NSynPerConn{n_syn_conn}.json"
    )  # Model with exactly <n_syn_conn> syn/conn (constant) fo all pathways
    nsynconn_model = model_types.AbstractModel.model_from_file(nsynconn_model_file)
    delay_model_file = os.path.join(
        TEST_DATA_DIR, f"model_config__DistDepDelay.json"
    )  # Deterministic delay model w/o variation
    delay_model = model_types.AbstractModel.model_from_file(delay_model_file)
    pct = 100.0

    # Case 1: Check connectivity with conn. prob. p=0.0 (no connectivity)
    prob_model_file = os.path.join(TEST_DATA_DIR, "model_config__ConnProb0p0.json")
    prob_model = model_types.AbstractModel.model_from_file(prob_model_file)

    ## (a) Empty edges table
    writer = EdgeWriter(None, edges_table_empty)
    manipulation(nodes, writer).apply(
        tgt_ids,
        amount_pct=pct,
        prob_model_spec={"file": prob_model_file},
        nsynconn_model_spec={"file": nsynconn_model_file},
    )
    res = writer.to_pandas()
    assert res.equals(edges_table_empty[res.columns]), "ERROR: Existing edges table changed!"

    ## (b) Edges already existing
    writer = EdgeWriter(None, edges_table)
    res = manipulation(nodes, writer).apply(
        tgt_ids,
        amount_pct=pct,
        prob_model_spec={"file": prob_model_file},
        nsynconn_model_spec={"file": nsynconn_model_file},
    )
    res = writer.to_pandas()
    assert res.equals(edges_table[res.columns]), "ERROR: Existing edges table changed!"

    ## (c) Standalone wiring per pathway
    pathway_nodes = nodes[0]
    assert nodes[0] is nodes[1]
    pathway_models = []
    for pre_mt in src_mtypes:
        for post_mt in tgt_mtypes:
            pathway_models.append(
                {
                    "pre": pre_mt,
                    "post": post_mt,
                    "prob_model": prob_model,
                    "nsynconn_model": nsynconn_model,
                    "delay_model": None,
                }
            )
    res = manipulation.connectome_wiring_per_pathway(nodes, pathway_models, seed=0, morph_ext="swc")
    assert res.size == 0, "ERROR: Connectome should be empty!"
    assert np.all(np.isin(required_properties, res.columns)), "ERROR: Synapse properties missing!"

    # Case 2: Check connectivity with conn. prob. p=1.0 (full connectivity, w/o autapses)
    prob_model_file = os.path.join(TEST_DATA_DIR, "model_config__ConnProb1p0.json")
    prob_model = model_types.AbstractModel.model_from_file(prob_model_file)

    ## (a) Empty edges table
    writer = EdgeWriter(None)
    manipulation(nodes, writer).apply(
        tgt_ids,
        amount_pct=pct,
        prob_model_spec={"file": prob_model_file},
        nsynconn_model_spec={"file": nsynconn_model_file},
    )
    res = writer.to_pandas()
    assert (
        res.shape[0]
        == (len(src_ids) * len(tgt_ids) - len(np.intersect1d(src_ids, tgt_ids))) * n_syn_conn
    ), "ERROR: Wrong number of synapses!"  # Check #synapses
    assert np.all(np.isin(required_properties, res.keys())), "ERROR: Synapse properties missing!"
    assert np.all(
        np.unique(res[["@source_node", "@target_node"]], axis=0, return_counts=True)[1]
        == n_syn_conn
    ), "ERROR: Wrong #syn/conn!"  # Check #synapses/connection

    for i in range(res.shape[0]):
        # Check synapse class consistency
        syn_cl = nodes[0].get(res.iloc[i]["@source_node"], properties="synapse_class").to_numpy()[0]
        if syn_cl == "EXC":
            assert res.iloc[i]["syn_type_id"] >= 100, "ERROR: Wrong EXC type ID!"
        elif syn_cl == "INH":
            assert res.iloc[i]["syn_type_id"] < 100, "ERROR: Wrong INH type ID!"
        else:
            assert False, "Synapse class unknown!"

        # Check synapse position/type consistency
        syn_pos = res.iloc[i][["afferent_center_x", "afferent_center_y", "afferent_center_z"]]
        sec_id, sec_pos, sec_type = res.iloc[i][
            ["afferent_section_id", "afferent_section_pos", "afferent_section_type"]
        ]
        if sec_id == 0:  # Soma section
            assert sec_pos == 0.0 and sec_type == SEC_SOMA, "ERROR: Soma section error!"
            assert np.all(
                np.isclose(
                    syn_pos.to_numpy(), nodes[1].positions(res.iloc[i]["@target_node"]).to_numpy()
                )
            ), "ERROR: Soma position error!"
        else:
            morph = get_tgt_morph(int(res.iloc[i]["@target_node"]))
            sec_id = int(
                sec_id - 1
            )  # IMPORTANT: Section IDs in NeuroM morphology don't include soma, so they need to be shifted by 1 (Soma ID is 0 in edges table)
            assert (
                sec_type == SEC_TYPE_MAP[morph.section(sec_id).type]
            ), "ERROR: Section type mismatch!"
            assert np.all(
                np.isclose(
                    nm.morphmath.path_fraction_point(morph.section(sec_id).points, sec_pos), syn_pos
                )
            ), "ERROR: Section position error!"

    ## (b) Edges already existing
    writer = EdgeWriter(None, edges_table)
    manipulation(nodes, writer).apply(
        tgt_ids,
        amount_pct=pct,
        prob_model_spec={"file": prob_model_file},
        nsynconn_model_spec={"file": nsynconn_model_file},
    )
    res = writer.to_pandas()
    assert (
        res.shape[0]
        == edges_table.shape[0]
        + (len(src_ids) * len(tgt_ids) - len(np.intersect1d(src_ids, tgt_ids))) * n_syn_conn
    ), "ERROR: Wrong number of synapses!"
    syn_props = list(res.keys())
    assert np.all(
        [
            np.sum(np.all(edges_table.iloc[i][syn_props] == res[syn_props], 1)) == 1
            for i in range(edges_table.shape[0])
        ]
    ), "ERROR: Existing synapses changed!"  # Check if all existing synapses still exist exactly once
    assert np.all(np.isin(required_properties, syn_props)), "ERROR: Synapse properties missing!"

    ## (c) Standalone wiring per pathway
    pathway_models = []
    for pre_mt in src_mtypes:
        for post_mt in tgt_mtypes:
            pathway_models.append(
                {
                    "pre": pre_mt,
                    "post": post_mt,
                    "prob_model": prob_model,
                    "nsynconn_model": nsynconn_model,
                    "delay_model": None,
                }
            )
    res = manipulation.connectome_wiring_per_pathway(nodes, pathway_models, seed=0, morph_ext="swc")
    assert (
        res.shape[0]
        == (len(src_ids) * len(tgt_ids) - len(np.intersect1d(src_ids, tgt_ids))) * n_syn_conn
    ), "ERROR: Wrong number of synapses!"  # Check #synapses
    assert np.all(np.isin(required_properties, res.columns)), "ERROR: Synapse properties missing!"
    assert np.all(
        np.unique(res[["@source_node", "@target_node"]], axis=0, return_counts=True)[1]
        == n_syn_conn
    ), "ERROR: Wrong #syn/conn!"  # Check #synapses/connection

    # Case 3: Check pct
    for pct in np.linspace(0, 100, 6):
        writer = EdgeWriter(None)
        manipulation(nodes, writer).apply(
            tgt_ids,
            amount_pct=pct.tolist(),
            prob_model_spec={"file": prob_model_file},
            nsynconn_model_spec={"file": nsynconn_model_file},
        )
        res = writer.to_pandas()
        assert (
            res.shape[0]
            == (
                len(src_ids) * len(tgt_ids) * pct / 100
                - len(np.intersect1d(src_ids, tgt_ids)) * pct / 100
            )
            * n_syn_conn
        ), f"ERROR: Wrong number of synapses! for pct ({pct})"  # Check #synapses

    # Case 4: Check src/tgt_sel
    pct = 100.0

    ## (a) Per synapse class
    for src_class in ["EXC", "INH"]:
        for tgt_class in ["EXC", "INH"]:
            sel_src = {"synapse_class": src_class}
            sel_dest = {"synapse_class": tgt_class}
            writer = EdgeWriter(None)
            manipulation(nodes, writer).apply(
                tgt_ids,
                sel_src=sel_src,
                sel_dest=sel_dest,
                amount_pct=pct,
                prob_model_spec={"file": prob_model_file},
                nsynconn_model_spec={"file": nsynconn_model_file},
            )
            res = writer.to_pandas()
            assert np.all(
                np.isin(res["@source_node"], nodes[0].ids(sel_src))
            ), "ERROR: Source selection error!"
            assert np.all(
                np.isin(res["@target_node"], nodes[1].ids(sel_dest))
            ), "ERROR: Target selection error!"
            assert (
                res.shape[0]
                == (
                    len(nodes[0].ids(sel_src)) * len(nodes[1].ids(sel_dest))
                    - len(np.intersect1d(nodes[0].ids(sel_src), nodes[1].ids(sel_dest)))
                )
                * n_syn_conn
            ), "ERROR: Wrong number of synapses!"  # Check #synapses

    ## (b) Per mtype
    for src_mt in src_mtypes:
        for tgt_mt in tgt_mtypes:
            sel_src = {"mtype": src_mt}
            sel_dest = {"mtype": tgt_mt}

            ### Integrated wiring
            writer = EdgeWriter(None)
            manipulation(nodes, writer).apply(
                tgt_ids,
                sel_src=sel_src,
                sel_dest=sel_dest,
                amount_pct=pct,
                prob_model_spec={"file": prob_model_file},
                nsynconn_model_spec={"file": nsynconn_model_file},
            )
            res = writer.to_pandas()
            assert np.all(
                np.isin(res["@source_node"], nodes[0].ids(sel_src))
            ), "ERROR: Source selection error!"
            assert np.all(
                np.isin(res["@target_node"], nodes[1].ids(sel_dest))
            ), "ERROR: Target selection error!"
            assert (
                res.shape[0]
                == (
                    len(nodes[0].ids(sel_src)) * len(nodes[1].ids(sel_dest))
                    - len(np.intersect1d(nodes[0].ids(sel_src), nodes[1].ids(sel_dest)))
                )
                * n_syn_conn
            ), "ERROR: Wrong number of synapses!"  # Check #synapses

            ### Standalone wiring per pathway
            pathway_models = [
                {
                    "pre": src_mt,
                    "post": tgt_mt,
                    "prob_model": prob_model,
                    "nsynconn_model": nsynconn_model,
                    "delay_model": None,
                }
            ]
            res = manipulation.connectome_wiring_per_pathway(
                nodes, pathway_models, seed=0, morph_ext="swc"
            )
            assert np.all(
                np.isin(res["@source_node"], nodes[0].ids(sel_src))
            ), "ERROR: Source selection error!"
            assert np.all(
                np.isin(res["@target_node"], nodes[1].ids(sel_dest))
            ), "ERROR: Target selection error!"
            assert (
                res.shape[0]
                == (
                    len(nodes[0].ids(sel_src)) * len(nodes[1].ids(sel_dest))
                    - len(np.intersect1d(nodes[0].ids(sel_src), nodes[1].ids(sel_dest)))
                )
                * n_syn_conn
            ), "ERROR: Wrong number of synapses!"  # Check #synapses

    # Case 5: Check block-based processing
    split_ids_list = [tgt_ids[: len(tgt_ids) >> 1], tgt_ids[len(tgt_ids) >> 1 :]]
    writer = EdgeWriter(None)
    for i_split, split_ids in enumerate(split_ids_list):
        print(split_ids)
        manipulation(nodes, writer, i_split, len(split_ids_list)).apply(
            split_ids,
            amount_pct=pct,
            prob_model_spec={"file": prob_model_file},
            nsynconn_model_spec={"file": nsynconn_model_file},
        )
    res = writer.to_pandas()
    assert (
        res.shape[0]
        == (len(src_ids) * len(tgt_ids) - len(np.intersect1d(src_ids, tgt_ids))) * n_syn_conn
    ), "ERROR: Wrong number of synapses!"  # Check #synapses

    # Case 6: Check delays (from PRE neuron (soma) to POST synapse position)
    def check_delay(nodes, delay_model, res):
        for i in range(res.shape[0]):
            delay_offset = delay_model.get_param_dict()["delay_mean_coeff_a"]
            delay_scale = delay_model.get_param_dict()["delay_mean_coeff_b"]
            src_pos = nodes[0].positions(res.iloc[i]["@source_node"]).to_numpy()
            syn_pos = res.iloc[i][
                ["afferent_center_x", "afferent_center_y", "afferent_center_z"]
            ].to_numpy()
            dist = np.sqrt(np.sum((src_pos - syn_pos) ** 2))
            delay = delay_scale * dist + delay_offset
            assert np.all(np.isclose(res.iloc[i]["delay"], delay)), "ERROR: Delay mismatch!"

    ## (a) Integrated wiring
    writer = EdgeWriter(None)
    manipulation(nodes, writer).apply(
        edges_table_empty,
        tgt_ids,
        amount_pct=pct,
        prob_model_spec={"file": prob_model_file},
        nsynconn_model_spec={"file": nsynconn_model_file},
        delay_model_spec={"file": delay_model_file},
    )
    res = writer.to_pandas()
    check_delay(nodes, delay_model, res)

    ## (b) Standalone wiring per pathway
    pathway_models = []
    for pre_mt in src_mtypes:
        for post_mt in tgt_mtypes:
            pathway_models.append(
                {
                    "pre": pre_mt,
                    "post": post_mt,
                    "prob_model": prob_model,
                    "nsynconn_model": nsynconn_model,
                    "delay_model": delay_model,
                }
            )
    res = manipulation.connectome_wiring_per_pathway(nodes, pathway_models, seed=0, morph_ext="swc")
    check_delay([pathway_nodes, pathway_nodes], delay_model, res)

    # Case 7: Check connectivity with conn. prob. p=0.1
    prob_model_file = os.path.join(TEST_DATA_DIR, "model_config__ConnProb0p1.json")
    prob_model = model_types.AbstractModel.model_from_file(prob_model_file)

    ## (a) Integrated wiring
    np.random.seed(0)
    syn_counts = []
    for rep in range(
        30
    ):  # Estimate synapse counts over N repetitions => May be increased if variation still to large
        writer = EdgeWriter(None)
        res = manipulation(nodes, writer).apply(
            tgt_ids,
            amount_pct=pct,
            prob_model_spec={"file": prob_model_file},
            nsynconn_model_spec={"file": nsynconn_model_file},
        )
        res = writer.to_pandas()
        syn_counts.append(res.shape[0])
    assert np.std(syn_counts) > 0, "ERROR: No variability over repetitions!"
    assert np.isclose(
        np.mean(syn_counts),
        (len(src_ids) * len(tgt_ids) - len(np.intersect1d(src_ids, tgt_ids))) * 0.1 * n_syn_conn,
        atol=1.5,
    ), f"ERROR: Wrong number of synapses!"  # Accept tolerance of +/-1.5

    ## (b) Standalone wiring per pathway [larger variability expected, since wiring per pathway]
    pathway_models = []
    for pre_mt in src_mtypes:
        for post_mt in tgt_mtypes:
            pathway_models.append(
                {
                    "pre": pre_mt,
                    "post": post_mt,
                    "prob_model": prob_model,
                    "nsynconn_model": nsynconn_model,
                    "delay_model": None,
                }
            )
    syn_counts = []
    for rep in range(
        40
    ):  # Estimate synapse counts over N repetitions => May be increased if variation still to large
        res = manipulation.connectome_wiring_per_pathway(
            nodes, pathway_models, seed=rep, morph_ext="swc"
        )
        syn_counts.append(res.shape[0])
    assert np.std(syn_counts) > 0, "ERROR: No variability over repetitions!"
    assert np.isclose(
        np.mean(syn_counts),
        (len(src_ids) * len(tgt_ids) - len(np.intersect1d(src_ids, tgt_ids))) * 0.1 * n_syn_conn,
        atol=1.5,
    ), f"ERROR: Wrong number of synapses!"  # Accept tolerance of +/-1.5



#!/usr/bin/env python

# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

import importlib.util
from setuptools import setup, find_packages

# read the contents of the README file
with open("README.rst", encoding="utf-8") as f:
    README = f.read()

spec = importlib.util.spec_from_file_location(
    "connectome_manipulator.version",
    "connectome_manipulator/version.py",
)
module = importlib.util.module_from_spec(spec)
spec.loader.exec_module(module)
VERSION = module.__version__


setup(
    name="connectome-manipulator",
    author="Blue Brain Project, EPFL",
    version=VERSION,
    description="A connectome manipulation framework for SONATA circuits",
    long_description=README,
    long_description_content_type="text/x-rst",
    url="https://github.com/BlueBrain/connectome-manipulator.git",
    license="Apache-2",
    install_requires=[
        "bluepysnap>=3.0.1",
        "numpy>=1.24.3",
        "pandas>=1.5.3",
        "progressbar>=2.5",
        "pyarrow>=10.0.1",
        "scipy>=1.10.1",
        "scikit-learn>=1.3.2",
        "voxcell>=3.1.5",
        "tables>=3.8.0",  # Optional dependency of pandas.DataFrame.to_hdf()
        "distributed>=2023.4.1",  # Dask
        "dask-mpi>=2022.4.0",
    ],
    packages=find_packages(),
    python_requires=">=3.10",
    extras_require={"docs": ["sphinx", "sphinx-bluebrain-theme"]},
    entry_points={
        "console_scripts": [
            "connectome-manipulator=connectome_manipulator.cli:app",
            "parallel-manipulator=connectome_manipulator.cli_parallel:app",
        ]
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
)



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

import connectome_manipulator.connectome_comparison.properties as test_module


def test_null():
    pass



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""Connectome wiring module (specialized)."""

from datetime import datetime, timedelta

import libsonata
import neurom as nm
import numpy as np
import tqdm

from connectome_manipulator import log, profiler
from connectome_manipulator.access_functions import (
    get_attribute,
    get_node_ids,
    get_enumeration,
    get_node_positions,
)
from connectome_manipulator.connectome_manipulation.converters import EdgeWriter
from connectome_manipulator.connectome_manipulation.manipulation import (
    MorphologyCachingManipulation,
)
from connectome_manipulator.model_building import model_types, conn_prob

# IDEAs for improvements:
#   Add model for synapse placement


class ConnectomeWiring(MorphologyCachingManipulation):
    """Special case of connectome manipulation class for wiring a connectome:

    Special operation of connectome wiring, which generates an empty connectome
    from scratch, or simply adds connections to an existing connectome.
    The manipulation can be applied through the :func:`apply` method.

    IMPORTANT: This is a highly optimized operation for wiring huge connectomes by pathways
    (i.e., for each pair of pre-/post-synaptic hemisphere/region/m-type). However, only
    specific properties like source/target node, afferent synapse positions, synapse type
    (INH: 0, EXC: 100), and delay (optional) will be generated.
    For general purpose wiring, please use the
    :func:`connectome_manipulator.connectome_manipulation.manipulation.conn_rewiring.ConnectomeRewiring`
    operation!
    """

    # SONATA section type mapping (as in MorphIO): 1 = soma, 2 = axon, 3 = basal, 4 = apical
    SEC_SOMA = 1
    SEC_TYPE_MAP = {nm.AXON: 2, nm.BASAL_DENDRITE: 3, nm.APICAL_DENDRITE: 4}

    @profiler.profileit(name="conn_wiring")
    def apply(
        self,
        split_ids,
        sel_src=None,
        sel_dest=None,
        pos_map_file=None,
        amount_pct=100.0,
        morph_ext="swc",
        prob_model_spec=None,
        nsynconn_model_spec=None,
        delay_model_spec=None,
        pathway_specs=None,
        **kwargs,
    ):
        """Applies a wiring (generation) of structural connections between pairs of neurons based on a given connectivity model.

        Args:
            split_ids (list-like): List of neuron IDs that are part of the current data split; will be automatically provided by the manipulator framework
            sel_src (str/list-like/dict): Source (pre-synaptic) neuron selection
            sel_dest (str/list-like/dict): Target (post-synaptic) neuron selection
            pos_map_file (str/list-like): Optional position mapping file pointing to a position mapping model (.json) or voxel data map (.nrrd); one or two files for source/target node populations may be provided
            amount_pct (float): Percentage of randomly sampled target (post-synaptic) neurons that will be wired
            morph_ext (str): Morphology file extension, e.g., "swc", "asc", "h5"
            prob_model_spec (dict): Connection probability model specification; a file can be specified by ``{"file": "path/file.json"}``
            nsynconn_model_spec (dict): Model specifications for #synapses/connection; a file can be specified by ``{"file": "path/file.json"}``
            delay_model_spec (dict): Delay model specification; a file can be specified by ``{"file": "path/file.json"}``
            pathway_specs (dict): Optional model specifications for efficiently setting model coefficients by pathway; will be automatically provided by the manipulator framework in case a .parquet file (containing a coefficient table for all pathways) is specified under "model_pathways" in the manipulation configuration file; only works with specific types of models
            **kwargs: Additional keyword arguments - Not used

        Note:
            Only structural synapse properties will be set: pre-/postsynaptic neuron IDs, synapse positions, type, axonal delays
        """
        # pylint: disable=arguments-differ
        assert len(kwargs) == 0
        if not prob_model_spec:
            prob_model_spec = {
                "model": "ConnProb1stOrderModel",
            }  # Default 1st-oder model
        if not nsynconn_model_spec:
            nsynconn_model_spec = {
                "model": "NSynConnModel",
            }  # Default #syn/conn model
        if not delay_model_spec:
            delay_model_spec = {
                "model": "LinDelayModel",
            }  # Default linear delay model
        for spec in (prob_model_spec, nsynconn_model_spec, delay_model_spec):
            # AbstractModel insists that "file" is the only key if present
            if "file" not in spec:
                spec["src_type_map"] = self.src_type_map
                spec["tgt_type_map"] = self.tgt_type_map
                spec["pathway_specs"] = pathway_specs
        # pylint: disable=arguments-differ, arguments-renamed
        log.log_assert(0.0 <= amount_pct <= 100.0, "amount_pct out of range!")

        with profiler.profileit(name="conn_wiring/setup"):
            # Intersect target nodes with split IDs and return if intersection is empty
            tgt_node_ids = get_node_ids(self.nodes[1], sel_dest, split_ids)
            num_tgt_total = len(tgt_node_ids)
            if num_tgt_total == 0:  # Nothing to wire
                log.info("No target nodes selected, nothing to wire")
                return
            if amount_pct < 100:
                num_tgt = np.round(amount_pct * num_tgt_total / 100).astype(int)
                tgt_sel = np.random.permutation(
                    np.concatenate(
                        (np.full(num_tgt, True), np.full(num_tgt_total - num_tgt, False)), axis=None
                    )
                )
            else:
                num_tgt = num_tgt_total
                tgt_sel = np.full(num_tgt_total, True)
            if num_tgt == 0:  # Nothing to wire
                log.info("No target nodes selected, nothing to wire")
                return
            # Load connection probability model
            p_model = model_types.AbstractModel.init_model(prob_model_spec)
            log.debug(f'Loaded conn. prob. model of type "{p_model.__class__.__name__}"')

            # Load #synapses/connection model
            nsynconn_model = model_types.AbstractModel.init_model(nsynconn_model_spec)
            log.debug(
                f'Loaded #synapses/connection model of type "{nsynconn_model.__class__.__name__}"'
            )

            # Load delay model (optional)
            if delay_model_spec is not None:
                delay_model = model_types.AbstractModel.init_model(delay_model_spec)
                log.debug(f'Loaded delay model of type "{delay_model.__class__.__name__}"')
            else:
                delay_model = None
                log.debug("No delay model provided")

            # Load source/taget position mappings (optional; two types of mappings supported)
            pos_mappings = conn_prob.get_pos_mapping_fcts(pos_map_file)

            # Determine source/target nodes for wiring
            src_node_ids = get_node_ids(self.nodes[0], sel_src)
            src_class = get_attribute(self.nodes[0], "synapse_class", src_node_ids)
            src_mtypes = get_enumeration(self.nodes[0], "mtype", src_node_ids)
            log.log_assert(len(src_node_ids) > 0, "No source nodes selected!")

            tgt_node_ids = tgt_node_ids[tgt_sel]  # Select subset of neurons (keeping order)
            tgt_mtypes = get_enumeration(self.nodes[1], "mtype", tgt_node_ids)

            # Get source/target node positions (optionally: two types of mappings)
            src_pos, tgt_pos = conn_prob.get_neuron_positions(
                self.nodes,
                [src_node_ids, tgt_node_ids],
                pos_acc=pos_mappings[0],
                vox_map=pos_mappings[1],
            )
            # ...and source positions w/o mapping (required for delays)
            if all(_map is None for _map in pos_mappings):
                raw_src_pos = src_pos
            else:
                raw_src_pos, _ = get_node_positions(self.nodes[0], src_node_ids)

            log.info(
                f"Generating afferent connections to {num_tgt} ({amount_pct}%) of {len(tgt_sel)} target neurons in current split (total={num_tgt_total}, sel_src={sel_src}, sel_dest={sel_dest})"
            )

        # Run connection wiring
        self._connectome_wiring_wrapper(
            src_node_ids,
            src_pos,
            src_mtypes,
            src_class,
            morph_ext,
            tgt_node_ids,
            tgt_pos,
            tgt_mtypes,
            p_model,
            nsynconn_model,
            delay_model,
            raw_src_pos,
        )

    @profiler.profileit(name="conn_wiring/wiring")
    def _connectome_wiring_wrapper(
        self,
        src_node_ids,
        src_positions,
        src_mtypes,
        src_class,
        morph_ext,
        tgt_node_ids,
        tgt_positions,
        tgt_mtypes,
        p_model,
        nsynconn_model,
        delay_model,
        raw_src_positions,  # src positions w/o pos mapping (for delays!)
    ):
        """Stand-alone wrapper for connectome wiring."""
        # get morphologies for this selection
        tgt_morphs = self._get_tgt_morphs(morph_ext, libsonata.Selection(tgt_node_ids))

        log_time = datetime.now()
        for tidx, (tgt, morph) in enumerate(zip(tgt_node_ids, tgt_morphs)):
            new_time = datetime.now()
            if (new_time - log_time) / timedelta(minutes=1) > 1:
                log.info(
                    "Processing target node %d out of %d",
                    tidx,
                    len(tgt_node_ids),
                )
                log_time = new_time

            # Determine conn. prob. of all source nodes to be connected with target node
            tgt_pos = tgt_positions[tidx : tidx + 1, :]
            p_src = p_model.apply(
                src_pos=src_positions,
                tgt_pos=tgt_pos,
                src_type=src_mtypes,
                tgt_type=[tgt_mtypes[tidx]],
                src_nid=src_node_ids,
                tgt_nid=[tgt],
            ).flatten()
            p_src[np.isnan(p_src)] = 0.0  # Exclude invalid values
            # Exclude autapses [ASSUMING node IDs are unique across src/tgt
            # node populations!]
            p_src[src_node_ids == tgt] = 0.0

            # Sample new presynaptic neurons from list of source nodes according to conn. prob.
            src_new_sel = np.random.rand(len(src_node_ids)) < p_src
            src_new = src_node_ids[src_new_sel]  # New source node IDs per connection
            num_new = len(src_new)
            if num_new == 0:
                continue  # Nothing to wire

            # Sample number of synapses per connection (mtype-specific)
            num_syn_per_conn = nsynconn_model.apply(
                src_type=src_mtypes[src_new_sel], tgt_type=tgt_mtypes[tidx]
            )
            syn_conn_idx = np.concatenate(
                [[i] * n for i, n in enumerate(num_syn_per_conn)]
            )  # Create mapping from synapses to connections
            num_gen_syn = len(syn_conn_idx)  # Number of synapses to generate

            # Get available dendritic sections (plus soma) to place synapses on
            sec_ind = np.hstack(
                [
                    [-1],  # [Soma]
                    np.flatnonzero(
                        np.isin(morph.section_types, [nm.BASAL_DENDRITE, nm.APICAL_DENDRITE])
                    ),
                ]
            )

            # Randomly choose section indices
            sec_sel = np.random.choice(sec_ind, len(syn_conn_idx))

            # Randomly choose fractional offset within each section
            off_sel = np.random.rand(len(syn_conn_idx))
            off_sel[sec_sel == -1] = 0.0  # Soma offsets must be zero

            # Synapse positions & (mapped) section types, computed from section & offset
            type_sel = np.full_like(sec_sel, self.SEC_SOMA)
            pos_sel = np.tile(morph.soma.center.astype(float), (len(sec_sel), 1))
            for idx in np.flatnonzero(sec_sel >= 0):
                type_sel[idx] = self.SEC_TYPE_MAP[morph.section(sec_sel[idx]).type]
                pos_sel[idx] = nm.morphmath.path_fraction_point(
                    morph.section(sec_sel[idx]).points, off_sel[idx]
                )

            # Synapse type assignment [INH: 0-99 (Using 0); EXC: >=100 (Using 100)]
            syn_type = np.select(
                [
                    src_class[src_new_sel][syn_conn_idx] == "INH",
                    src_class[src_new_sel][syn_conn_idx] == "EXC",
                ],
                [np.full(num_gen_syn, 0), np.full(num_gen_syn, 100)],
            )

            # Assign distance-dependent delays (mtype-specific), based on (generative) delay model (optional)
            # IMPORTANT: Distances for delays are computed in them original coordinate system w/o coordinate transformation!
            kwargs = {}
            if delay_model is not None:
                src_new_pos = raw_src_positions[src_new_sel, :]
                syn_dist = np.sqrt(
                    np.sum((pos_sel - src_new_pos[syn_conn_idx, :]) ** 2, 1)
                )  # Distance from source neurons (soma) to synapse positions on target neuron
                delay = delay_model.apply(
                    distance=syn_dist,
                    src_type=src_mtypes[src_new_sel][syn_conn_idx],
                    tgt_type=tgt_mtypes[tidx],
                )
                if np.isscalar(delay):
                    kwargs["delay"] = np.full(syn_type.shape, delay)
                else:
                    kwargs["delay"] = delay

            # IMPORTANT: Section IDs in NeuroM morphology don't include soma, so they need to be shifted by 1 (Soma ID is 0 in edges table)
            #            The tuple representation (section_id, offset) is used here.
            self.writer.append(
                source_node_id=src_new[syn_conn_idx],
                target_node_id=np.full_like(syn_type, tgt),
                afferent_section_id=sec_sel + 1,
                afferent_section_pos=off_sel,
                afferent_section_type=type_sel,
                afferent_center_x=pos_sel[:, 0],
                afferent_center_y=pos_sel[:, 1],
                afferent_center_z=pos_sel[:, 2],
                syn_type_id=syn_type,
                edge_type_id=np.zeros_like(syn_type),
                **kwargs,
            )

    @classmethod
    def connectome_wiring_per_pathway(cls, nodes, pathway_models, seed=0, morph_ext="h5"):
        """Stand-alone connectome wiring per pathway, i.e., wiring pathways using pathway-specific probability/nsynconn/delay models."""
        # Init random seed for connectome building and sampling from parameter distributions
        np.random.seed(seed)

        with_delay = any(d["delay_model"] for d in pathway_models)

        writer = EdgeWriter(None, with_delay=with_delay)
        conn_wiring = cls(nodes, writer)
        src_nodes, tgt_nodes = nodes

        # Loop over pathways
        for pathway_dict in tqdm.tqdm(pathway_models):
            # [OPTIMIZATION: Run wiring of pathways in parallel]

            pre_type = pathway_dict["pre"]
            post_type = pathway_dict["post"]
            prob_model = pathway_dict["prob_model"]
            nsynconn_model = pathway_dict["nsynconn_model"]
            delay_model = pathway_dict["delay_model"]

            # Select source/target nodes
            src_node_ids = src_nodes.ids({"mtype": pre_type})
            src_class = get_attribute(src_nodes, "synapse_class", src_node_ids)
            src_mtypes = get_enumeration(src_nodes, "mtype", src_node_ids)
            src_positions = src_nodes.positions(
                src_node_ids
            ).to_numpy()  # OPTIONAL: Coordinate system transformation may be added here

            tgt_node_ids = tgt_nodes.ids({"mtype": post_type})
            tgt_mtypes = get_enumeration(tgt_nodes, "mtype", tgt_node_ids)
            tgt_positions = tgt_nodes.positions(
                tgt_node_ids
            ).to_numpy()  # OPTIONAL: Coordinate system transformation may be added here

            # Create edges per pathway
            # pylint: disable=protected-access
            conn_wiring._connectome_wiring_wrapper(
                src_node_ids,
                src_positions,
                src_mtypes,
                src_class,
                morph_ext,
                tgt_node_ids,
                tgt_positions,
                tgt_mtypes,
                prob_model,
                nsynconn_model,
                delay_model,
                src_positions,
            )

            # ALTERNATIVE: Write to .parquet file and merge/convert to SONATA later
            # ... connectome_manipulation.edges_to_parquet(edges_table, output_file)
            # ... connectome_manipulation.parquet_to_sonata(input_file_list, output_file, nodes, nodes_files, keep_parquet=False)

        # Merge edges, re-sort, and assign new index
        edges_table = writer.to_pandas()
        edges_table.sort_values(["@target_node", "@source_node"], inplace=True)
        edges_table.reset_index(inplace=True, drop=True)

        return edges_table



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""Various utility functions"""

import json
import os
from pathlib import Path
from typing import Any, Dict

import numpy as np


def _apply_recursively(func, obj, origin=(0, 0, 0)):
    origin = np.array(origin)

    if hasattr(obj, "soma"):
        obj.soma.points = origin + func(obj.soma.points - origin)
    for s in obj.iter():
        s.points = origin + func(s.points - origin)


def transform(obj, A):
    """Apply transformation matrix `A` to a given morphology object.

    Note:
        This comes out of bluepysnap

    Args:
        obj: Morphology / Section
        A: rotation matrix (4 x 4 NumPy array)
    """
    if A is None:
        return
    A = np.asarray(A)
    if A.shape != (4, 4):
        raise ValueError(f"`A` should be 4 x 4 matrix (got: {A.shape})")
    A = A.transpose()

    def func(p):
        return np.dot(np.column_stack((p, np.ones(len(p)))), A)[:, :3]

    _apply_recursively(func, obj)


def create_dir(path: os.PathLike) -> Path:
    """Create directory and parents if it doesn't already exist."""
    path = Path(path)
    path.mkdir(parents=True, exist_ok=True)
    return path


def load_json(filepath: os.PathLike) -> Dict[Any, Any]:
    """Load from JSON file."""
    return json.loads(Path(filepath).read_bytes())


def write_json(filepath: os.PathLike, data: Dict[Any, Any], indent: int = 2) -> None:
    """Write json file."""
    with open(filepath, "w", encoding="utf-8") as fd:
        json.dump(data, fd, indent=indent)


def reduce_config_paths(config: dict, config_dir: os.PathLike) -> dict:
    """Reduces absolute paths with respect to base_dir.

    Args:
        config (dict): Input config with absolute paths and no manifest.
        config_dir (str|Path): The path to the directory of the config.

    Returns:
        A new config with the paths reduced to being relative to the config dir.

    Note: Paths that are not in config_dir's tree will stay absolute.
    """
    if not Path(config_dir).is_absolute():
        raise ValueError(f"Circuit config's directory is not absolute: {config_dir}")

    if "manifest" in config:
        raise ValueError(
            f"A reduced config with absolute paths and no manifest must be provided.\n"
            f"Got instead: {config}"
        )

    reduced_config = {
        "version": config.get("version", "1"),
        "manifest": {"$BASE_DIR": "."},
    }

    if "node_sets_file" in config:
        reduced_config["node_sets_file"] = _reduce_path(config["node_sets_file"], config_dir)

    if "components" in config:
        # Reduce components dict recursively (may have nested dicts!)
        reduced_config["components"] = _reduce_dict(config["components"], config_dir)

    reducer = {
        "nodes_file": _reduce_path,
        "node_types_file": _reduce_path,
        "edges_file": _reduce_path,
        "edge_types_file": _reduce_path,
        "populations": _reduce_populations,
    }

    reduced_config["networks"] = {
        network_type: [
            {
                key: (reducer[key](value, config_dir) if key in reducer else value)
                for key, value in net_dict.items()
            }
            for net_dict in network_list
        ]
        for network_type, network_list in config["networks"].items()
    }
    return reduced_config


def _reduce_path(path: str, base_dir: os.PathLike) -> str:
    if not path or path.startswith("$"):
        return path

    path = Path(path)

    if path.is_absolute():
        if path.is_relative_to(base_dir):
            return str(Path("$BASE_DIR", path.relative_to(base_dir)))

    return str(Path("$BASE_DIR", path))


def _reduce_populations(populations_dict: dict, base_dir: os.PathLike) -> dict:
    def reduce_entry(key, value):
        if key.endswith(("file", "dir")):
            return _reduce_path(value, base_dir)

        if key == "alternate_morphologies":
            return {
                alt_key: _reduce_path(alt_path, base_dir) for alt_key, alt_path in value.items()
            }

        return value

    return {
        pop_name: {key: reduce_entry(key, value) for key, value in pop_dict.items()}
        for pop_name, pop_dict in populations_dict.items()
    }


def _reduce_dict(entry: dict, base_dir: os.PathLike) -> dict:
    reduced_dict = {}
    for key, value in entry.items():
        if isinstance(value, dict):
            reduced_dict[key] = _reduce_dict(value, base_dir)
        else:
            reduced_dict[key] = _reduce_path(value, base_dir)
    return reduced_dict


class ConsoleColors:
    """Helper class for formatting console text."""

    BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE, _, DEFAULT = range(30, 40)
    NORMAL, BOLD, DIM, UNDERLINED, BLINK, INVERTED, HIDDEN = [a << 8 for a in range(7)]

    # These are the sequences needed to control output
    _CHANGE_SEQ = "\033[{}m"
    _RESET_SEQ = "\033[0m"

    @classmethod
    def reset(cls):
        """Reset colors."""
        return cls._RESET_SEQ

    @classmethod
    def set_text_color(cls, color):
        """Change text color."""
        return cls._CHANGE_SEQ.format(color)

    @classmethod
    def format_text(cls, text, color, style=None):
        """Format the text."""
        style = (style or color) >> 8
        format_seq = str(color & 0x00FF) + ((";" + str(style)) if style else "")
        return cls._CHANGE_SEQ.format(format_seq) + text + cls._RESET_SEQ
